# pylint: disable=too-many-lines
import json
from textwrap import dedent
from threading import Lock
from unittest import TestCase

from pcs.common.str_tools import indent

from pcs_test.tier1.cib_resource.common import (
    ResourceTest,
    fixture_use_meta_command_instead_warning,
)
from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.bin_mock import get_mock_settings
from pcs_test.tools.fixture_cib import CachedCibFixture
from pcs_test.tools.misc import (
    ParametrizedTestMetaClass,
    get_tmp_file,
    is_minimum_pacemaker_version,
    outdent,
    skip_unless_crm_rule,
    write_data_to_tmpfile,
    write_file_to_tmpfile,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.pcs_runner import PcsRunner

PCMK_2_0_3_PLUS = is_minimum_pacemaker_version(2, 0, 3)
ERRORS_HAVE_OCCURRED = (
    "Error: Errors have occurred, therefore pcs is unable to continue\n"
)

empty_cib = rc("cib-empty.xml")


class StonithDescribeTest(TestCase, AssertPcsMixin):
    def setUp(self):
        self.pcs_runner = PcsRunner(cib_file=None)
        self.pcs_runner.mock_settings = get_mock_settings()

    def test_success(self):
        self.assert_pcs_success(
            "stonith describe fence_pcsmock_params".split(),
            stdout_start=dedent(
                """\
                fence_pcsmock_params - Mock agent for pcs tests - agent with various params

                This is an agent with params for pcs tests

                Stonith options:
                """
            ),
        )

    def test_full(self):
        self.assert_pcs_success(
            "stonith describe fence_pcsmock_params --full".split(),
            stdout_regexp=".*pcmk_list_retries.*",
        )

    def test_nonextisting_agent(self):
        self.assert_pcs_fail(
            "stonith describe fence_noexist".split(),
            (
                "Error: Agent 'stonith:fence_noexist' is not installed or does not "
                "provide valid metadata: "
                "pcs mock error message: unable to load agent metadata\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )

    def test_not_enough_params(self):
        self.assert_pcs_fail(
            "stonith describe".split(),
            stderr_start="\nUsage: pcs stonith describe...\n",
        )

    def test_too_many_params(self):
        self.assert_pcs_fail(
            "stonith describe agent1 agent2".split(),
            stderr_start="\nUsage: pcs stonith describe...\n",
        )

    def test_pcsd_interface(self):
        self.maxDiff = None
        stdout, stderr, returncode = self.pcs_runner.run(
            "stonith get_fence_agent_info stonith:fence_pcsmock_params".split()
        )
        self.assertEqual(
            json.loads(stdout),
            {
                "name": "stonith:fence_pcsmock_params",
                "standard": "stonith",
                "provider": None,
                "type": "fence_pcsmock_params",
                "shortdesc": "Mock agent for pcs tests - agent with various params",
                "longdesc": "This is an agent with params for pcs tests",
                "parameters": [
                    {
                        "name": "action",
                        "shortdesc": "Fencing action",
                        "longdesc": None,
                        "type": "string",
                        "default": "reboot",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": True,
                        "deprecated_by": [
                            "pcmk_off_action",
                            "pcmk_reboot_action",
                        ],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "ip",
                        "shortdesc": "IP address or hostname of fencing device",
                        "longdesc": None,
                        "type": "string",
                        "default": None,
                        "enum_values": None,
                        "required": True,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "ipaddr",
                        "shortdesc": "IP address or hostname of fencing device",
                        "longdesc": None,
                        "type": "string",
                        "default": None,
                        "enum_values": None,
                        "required": True,
                        "advanced": False,
                        "deprecated": True,
                        "deprecated_by": ["ip"],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "login",
                        "shortdesc": "Login name",
                        "longdesc": None,
                        "type": "string",
                        "default": None,
                        "enum_values": None,
                        "required": True,
                        "advanced": False,
                        "deprecated": True,
                        "deprecated_by": ["username"],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "passwd",
                        "shortdesc": "Login password or passphrase",
                        "longdesc": None,
                        "type": "string",
                        "default": None,
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": True,
                        "deprecated_by": ["password"],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "password",
                        "shortdesc": "Login password or passphrase",
                        "longdesc": None,
                        "type": "string",
                        "default": None,
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "secure",
                        "shortdesc": "Use SSH connection",
                        "longdesc": None,
                        "type": "boolean",
                        "default": None,
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": True,
                        "deprecated_by": ["ssh"],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "ssh",
                        "shortdesc": "Use SSH connection",
                        "longdesc": None,
                        "type": "boolean",
                        "default": None,
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "username",
                        "shortdesc": "Login name",
                        "longdesc": None,
                        "type": "string",
                        "default": None,
                        "enum_values": None,
                        "required": True,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "verbose",
                        "shortdesc": "Verbose mode",
                        "longdesc": None,
                        "type": "boolean",
                        "default": None,
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "debug",
                        "shortdesc": "Write debug information to given file",
                        "longdesc": None,
                        "type": "string",
                        "default": None,
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": True,
                        "deprecated_by": ["debug_file"],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "debug_file",
                        "shortdesc": "Write debug information to given file",
                        "longdesc": None,
                        "type": "string",
                        "default": None,
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_host_argument",
                        "shortdesc": "An alternate parameter to supply instead of 'port'",
                        "longdesc": "An alternate parameter to supply instead of 'port'.\nSome devices do not support the standard 'port' parameter or may provide additional ones. Use this to specify an alternate, device-specific, parameter that should indicate the machine to be fenced. A value of 'none' can be used to tell the cluster not to supply any additional parameters.",
                        "type": "string",
                        "default": "port",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_host_map",
                        "shortdesc": "A mapping of host names to ports numbers for devices that do not support host names.",
                        "longdesc": "A mapping of host names to ports numbers for devices that do not support host names.\nEg. node1:1;node2:2,3 would tell the cluster to use port 1 for node1 and ports 2 and 3 for node2",
                        "type": "string",
                        "default": "",
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_host_list",
                        "shortdesc": "A list of machines controlled by this device (Optional unless pcmk_host_check=static-list).",
                        "longdesc": "A list of machines controlled by this device (Optional unless pcmk_host_check=static-list).\nEg. node1,node2,node3",
                        "type": "string",
                        "default": "",
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_host_check",
                        "shortdesc": "How to determine which machines are controlled by the device.",
                        "longdesc": "How to determine which machines are controlled by the device.\nAllowed values: dynamic-list (query the device via the 'list' command), static-list (check the pcmk_host_list attribute), status (query the device via the 'status' command), none (assume every device can fence every machine)",
                        "type": "string",
                        "default": "dynamic-list",
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_delay_max",
                        "shortdesc": "Enable a delay of no more than the time specified before executing fencing actions. Pacemaker derives the overall delay by taking the value of pcmk_delay_base and adding a random delay value such that the sum is kept below this maximum.",
                        "longdesc": "Enable a delay of no more than the time specified before executing fencing actions. Pacemaker derives the overall delay by taking the value of pcmk_delay_base and adding a random delay value such that the sum is kept below this maximum.\nThis prevents double fencing when using slow devices such as sbd.\nUse this to enable a random delay for fencing actions.\nThe overall delay is derived from this random delay value adding a static delay so that the sum is kept below the maximum delay.",
                        "type": "time",
                        "default": "0s",
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_delay_base",
                        "shortdesc": "Enable a base delay for fencing actions and specify base delay value.",
                        "longdesc": 'Enable a base delay for fencing actions and specify base delay value.\nThis enables a static delay for fencing actions, which can help avoid "death matches" where two nodes try to fence each other at the same time. If pcmk_delay_max is also used, a random delay will be added such that the total delay is kept below that value.\nThis can be set to a single time value to apply to any node targeted by this device (useful if a separate device is configured for each target), or to a node map (for example, "node1:1s;node2:5") to set a different value per target.',
                        "type": "string",
                        "default": "0s",
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_action_limit",
                        "shortdesc": "The maximum number of actions can be performed in parallel on this device",
                        "longdesc": "The maximum number of actions can be performed in parallel on this device.\nCluster property concurrent-fencing=true needs to be configured first.\nThen use this to specify the maximum number of actions can be performed in parallel on this device. -1 is unlimited.",
                        "type": "integer",
                        "default": "1",
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_reboot_action",
                        "shortdesc": "An alternate command to run instead of 'reboot'",
                        "longdesc": "An alternate command to run instead of 'reboot'.\nSome devices do not support the standard commands or may provide additional ones.\nUse this to specify an alternate, device-specific, command that implements the 'reboot' action.",
                        "type": "string",
                        "default": "reboot",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_reboot_timeout",
                        "shortdesc": "Specify an alternate timeout to use for reboot actions instead of stonith-timeout",
                        "longdesc": "Specify an alternate timeout to use for reboot actions instead of stonith-timeout.\nSome devices need much more/less time to complete than normal.\nUse this to specify an alternate, device-specific, timeout for 'reboot' actions.",
                        "type": "time",
                        "default": "60s",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_reboot_retries",
                        "shortdesc": "The maximum number of times to retry the 'reboot' command within the timeout period",
                        "longdesc": "The maximum number of times to retry the 'reboot' command within the timeout period.\nSome devices do not support multiple connections. Operations may 'fail' if the device is busy with another task so Pacemaker will automatically retry the operation, if there is time remaining. Use this option to alter the number of times Pacemaker retries 'reboot' actions before giving up.",
                        "type": "integer",
                        "default": "2",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_off_action",
                        "shortdesc": "An alternate command to run instead of 'off'",
                        "longdesc": "An alternate command to run instead of 'off'.\nSome devices do not support the standard commands or may provide additional ones.\nUse this to specify an alternate, device-specific, command that implements the 'off' action.",
                        "type": "string",
                        "default": "off",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_off_timeout",
                        "shortdesc": "Specify an alternate timeout to use for off actions instead of stonith-timeout",
                        "longdesc": "Specify an alternate timeout to use for off actions instead of stonith-timeout.\nSome devices need much more/less time to complete than normal.\nUse this to specify an alternate, device-specific, timeout for 'off' actions.",
                        "type": "time",
                        "default": "60s",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_off_retries",
                        "shortdesc": "The maximum number of times to retry the 'off' command within the timeout period",
                        "longdesc": "The maximum number of times to retry the 'off' command within the timeout period.\nSome devices do not support multiple connections. Operations may 'fail' if the device is busy with another task so Pacemaker will automatically retry the operation, if there is time remaining. Use this option to alter the number of times Pacemaker retries 'off' actions before giving up.",
                        "type": "integer",
                        "default": "2",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_on_action",
                        "shortdesc": "An alternate command to run instead of 'on'",
                        "longdesc": "An alternate command to run instead of 'on'.\nSome devices do not support the standard commands or may provide additional ones.\nUse this to specify an alternate, device-specific, command that implements the 'on' action.",
                        "type": "string",
                        "default": "on",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_on_timeout",
                        "shortdesc": "Specify an alternate timeout to use for on actions instead of stonith-timeout",
                        "longdesc": "Specify an alternate timeout to use for on actions instead of stonith-timeout.\nSome devices need much more/less time to complete than normal.\nUse this to specify an alternate, device-specific, timeout for 'on' actions.",
                        "type": "time",
                        "default": "60s",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_on_retries",
                        "shortdesc": "The maximum number of times to retry the 'on' command within the timeout period",
                        "longdesc": "The maximum number of times to retry the 'on' command within the timeout period.\nSome devices do not support multiple connections. Operations may 'fail' if the device is busy with another task so Pacemaker will automatically retry the operation, if there is time remaining. Use this option to alter the number of times Pacemaker retries 'on' actions before giving up.",
                        "type": "integer",
                        "default": "2",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_list_action",
                        "shortdesc": "An alternate command to run instead of 'list'",
                        "longdesc": "An alternate command to run instead of 'list'.\nSome devices do not support the standard commands or may provide additional ones.\nUse this to specify an alternate, device-specific, command that implements the 'list' action.",
                        "type": "string",
                        "default": "list",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_list_timeout",
                        "shortdesc": "Specify an alternate timeout to use for list actions instead of stonith-timeout",
                        "longdesc": "Specify an alternate timeout to use for list actions instead of stonith-timeout.\nSome devices need much more/less time to complete than normal.\nUse this to specify an alternate, device-specific, timeout for 'list' actions.",
                        "type": "time",
                        "default": "60s",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_list_retries",
                        "shortdesc": "The maximum number of times to retry the 'list' command within the timeout period",
                        "longdesc": "The maximum number of times to retry the 'list' command within the timeout period.\nSome devices do not support multiple connections. Operations may 'fail' if the device is busy with another task so Pacemaker will automatically retry the operation, if there is time remaining. Use this option to alter the number of times Pacemaker retries 'list' actions before giving up.",
                        "type": "integer",
                        "default": "2",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_monitor_action",
                        "shortdesc": "An alternate command to run instead of 'monitor'",
                        "longdesc": "An alternate command to run instead of 'monitor'.\nSome devices do not support the standard commands or may provide additional ones.\nUse this to specify an alternate, device-specific, command that implements the 'monitor' action.",
                        "type": "string",
                        "default": "monitor",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_monitor_timeout",
                        "shortdesc": "Specify an alternate timeout to use for monitor actions instead of stonith-timeout",
                        "longdesc": "Specify an alternate timeout to use for monitor actions instead of stonith-timeout.\nSome devices need much more/less time to complete than normal.\nUse this to specify an alternate, device-specific, timeout for 'monitor' actions.",
                        "type": "time",
                        "default": "60s",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_monitor_retries",
                        "shortdesc": "The maximum number of times to retry the 'monitor' command within the timeout period",
                        "longdesc": "The maximum number of times to retry the 'monitor' command within the timeout period.\nSome devices do not support multiple connections. Operations may 'fail' if the device is busy with another task so Pacemaker will automatically retry the operation, if there is time remaining. Use this option to alter the number of times Pacemaker retries 'monitor' actions before giving up.",
                        "type": "integer",
                        "default": "2",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_status_action",
                        "shortdesc": "An alternate command to run instead of 'status'",
                        "longdesc": "An alternate command to run instead of 'status'.\nSome devices do not support the standard commands or may provide additional ones.\nUse this to specify an alternate, device-specific, command that implements the 'status' action.",
                        "type": "string",
                        "default": "status",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_status_timeout",
                        "shortdesc": "Specify an alternate timeout to use for status actions instead of stonith-timeout",
                        "longdesc": "Specify an alternate timeout to use for status actions instead of stonith-timeout.\nSome devices need much more/less time to complete than normal.\nUse this to specify an alternate, device-specific, timeout for 'status' actions.",
                        "type": "time",
                        "default": "60s",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                    {
                        "name": "pcmk_status_retries",
                        "shortdesc": "The maximum number of times to retry the 'status' command within the timeout period",
                        "longdesc": "The maximum number of times to retry the 'status' command within the timeout period.\nSome devices do not support multiple connections. Operations may 'fail' if the device is busy with another task so Pacemaker will automatically retry the operation, if there is time remaining. Use this option to alter the number of times Pacemaker retries 'status' actions before giving up.",
                        "type": "integer",
                        "default": "2",
                        "enum_values": None,
                        "required": False,
                        "advanced": True,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                ],
                "actions": [
                    {
                        "name": "on",
                        "timeout": None,
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "off",
                        "timeout": None,
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "reboot",
                        "timeout": None,
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "metadata",
                        "timeout": None,
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "status",
                        "timeout": None,
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "monitor",
                        "timeout": None,
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "list",
                        "timeout": None,
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "list-status",
                        "timeout": None,
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "stop",
                        "timeout": "20s",
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "start",
                        "timeout": "20s",
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "validate-all",
                        "timeout": None,
                        "interval": None,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                ],
                "default_actions": [
                    {
                        "name": "monitor",
                        "timeout": None,
                        "interval": "60s",
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    }
                ],
            },
        )
        self.assertEqual(stderr, "")
        self.assertEqual(returncode, 0)


class StonithTest(TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_test_stonith")
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        self.temp_corosync_conf = get_tmp_file("tier1_test_stonith")
        write_file_to_tmpfile(rc("corosync.conf"), self.temp_corosync_conf)
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.pcs_runner.mock_settings = get_mock_settings()
        self.pcs_runner.mock_settings["corosync_conf_file"] = (
            self.temp_corosync_conf.name
        )

    def tearDown(self):
        self.temp_cib.close()
        self.temp_corosync_conf.close()

    def test_stonith_creation_nonexistent_agent(self):
        self.assert_pcs_fail(
            "stonith create test1 fence_noexist".split(),
            (
                "Error: Agent 'stonith:fence_noexist' is not installed or does not "
                "provide valid metadata: "
                "pcs mock error message: unable to load agent metadata, "
                "use --force to override\n" + ERRORS_HAVE_OCCURRED
            ),
        )

        self.assert_pcs_success(
            "stonith create test1 fence_noexist --force".split(),
            stderr_full=(
                "Warning: Agent 'stonith:fence_noexist' is not installed or does not "
                "provide valid metadata: "
                "pcs mock error message: unable to load agent metadata\n"
            ),
        )

        self.assert_pcs_success(
            "stonith config".split(),
            dedent(
                """\
                Resource: test1 (class=stonith type=fence_noexist)
                  Operations:
                    monitor: test1-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_creation_pcmk_status_action(self):
        self.assert_pcs_success(
            "stonith create test9 fence_pcsmock_minimal pcmk_status_action=xxx".split(),
        )

        self.assert_pcs_success(
            "stonith config test9".split(),
            dedent(
                """\
                Resource: test9 (class=stonith type=fence_pcsmock_minimal)
                  Attributes: test9-instance_attributes
                    pcmk_status_action=xxx
                  Operations:
                    monitor: test9-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_creation_pcmk_params(self):
        # Testing that pcmk_host_check, pcmk_host_list & pcmk_host_map are
        # allowed for stonith agents
        self.assert_pcs_success(
            (
                "stonith create fencing fence_pcsmock_minimal "
                "pcmk_host_map=buzz-01:1;buzz-02:2;buzz-03:3;buzz-04:4;buzz-05:5 "
                "pcmk_host_check=static-list "
                "pcmk_host_list=buzz-01,buzz-02,buzz-03,buzz-04,buzz-05"
            ).split(),
        )

        self.assert_pcs_fail(
            "resource config fencing".split(),
            (
                "Warning: Unable to find resource 'fencing'\n"
                "Error: No resource found\n"
            ),
        )

        self.assert_pcs_success(
            "stonith config fencing".split(),
            dedent(
                """\
                Resource: fencing (class=stonith type=fence_pcsmock_minimal)
                  Attributes: fencing-instance_attributes
                    pcmk_host_check=static-list
                    pcmk_host_list=buzz-01,buzz-02,buzz-03,buzz-04,buzz-05
                    pcmk_host_map=buzz-01:1;buzz-02:2;buzz-03:3;buzz-04:4;buzz-05:5
                  Operations:
                    monitor: fencing-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_creation_pcmk_host_list(self):
        self.assert_pcs_success(
            [
                "stonith",
                "create",
                "F1",
                "fence_pcsmock_minimal",
                "pcmk_host_list=nodea nodeb",
            ],
        )

        self.assert_pcs_success(
            "stonith config F1".split(),
            dedent(
                """\
                Resource: F1 (class=stonith type=fence_pcsmock_minimal)
                  Attributes: F1-instance_attributes
                    pcmk_host_list="nodea nodeb"
                  Operations:
                    monitor: F1-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_creation(self):
        self.assert_pcs_fail(
            "stonith create test2 fence_pcsmock_params".split(),
            (
                "Error: stonith option 'ip' or 'ipaddr' (deprecated) has to be "
                "specified, use --force to override\n"
                "Error: stonith option 'username' or 'login' (deprecated) has "
                "to be specified, use --force to override\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )

        self.assert_pcs_success(
            "stonith create test2 fence_pcsmock_params --force".split(),
            stderr_start=(
                "Warning: stonith option 'ip' or 'ipaddr' (deprecated) has to be "
                "specified\n"
                "Warning: stonith option 'username' or 'login' (deprecated) has to "
                "be specified\n"
            ),
        )

        self.assert_pcs_fail(
            "stonith create test3 fence_pcsmock_params bad_argument=test".split(),
            stderr_start=(
                "Error: invalid stonith option 'bad_argument', allowed options are:"
            ),
        )

        self.assert_pcs_fail(
            "stonith create test3 fence_pcsmock_params ip=test".split(),
            (
                "Error: stonith option 'username' or 'login' (deprecated) has "
                "to be specified, use --force to override\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )

        self.assert_pcs_success(
            "stonith create test3 fence_pcsmock_params ip=test --force".split(),
            stderr_start=(
                "Warning: stonith option 'username' or 'login' (deprecated) "
                "has to be specified\n"
            ),
        )

        self.assert_pcs_success(
            "stonith config test2".split(),
            dedent(
                """\
                Resource: test2 (class=stonith type=fence_pcsmock_params)
                  Operations:
                    monitor: test2-monitor-interval-60s
                      interval=60s
                """
            ),
        )

        self.assert_pcs_success(
            "stonith config".split(),
            dedent(
                """\
                Resource: test2 (class=stonith type=fence_pcsmock_params)
                  Operations:
                    monitor: test2-monitor-interval-60s
                      interval=60s
                Resource: test3 (class=stonith type=fence_pcsmock_params)
                  Attributes: test3-instance_attributes
                    ip=test
                  Operations:
                    monitor: test3-monitor-interval-60s
                      interval=60s
                """
            ),
        )

        self.assert_pcs_success(
            [
                "stonith",
                "create",
                "test-fencing",
                "fence_pcsmock_minimal",
                "pcmk_host_list=rhel7-node1 rhel7-node2",
                "op",
                "monitor",
                "interval=61s",
            ],
        )

        self.assert_pcs_success(
            "config".split(),
            dedent(
                """\
                Cluster Name: test99
                Corosync Nodes:
                 rh7-1 rh7-2
                Pacemaker Nodes:

                Stonith Devices:
                  Resource: test2 (class=stonith type=fence_pcsmock_params)
                    Operations:
                      monitor: test2-monitor-interval-60s
                        interval=60s
                  Resource: test3 (class=stonith type=fence_pcsmock_params)
                    Attributes: test3-instance_attributes
                      ip=test
                    Operations:
                      monitor: test3-monitor-interval-60s
                        interval=60s
                  Resource: test-fencing (class=stonith type=fence_pcsmock_minimal)
                    Attributes: test-fencing-instance_attributes
                      pcmk_host_list="rhel7-node1 rhel7-node2"
                    Operations:
                      monitor: test-fencing-monitor-interval-61s
                        interval=61s
                """
            ),
        )

    def test_stonith_agent_validation(self):
        self.pcs_runner.mock_settings = get_mock_settings()
        self.assert_pcs_fail(
            (
                "stonith create fencing fence_pcsmock_params "
                "ip=is_invalid=True username=apc "
                "--agent-validation"
            ).split(),
            stderr_full=(
                "Error: Validation result from agent (use --force to override):\n"
                "  pcsmock validation failure\n" + ERRORS_HAVE_OCCURRED
            ),
        )

        self.assert_pcs_success(
            (
                "stonith create fencing fence_pcsmock_params "
                "ip=is_invalid=True username=apc "
                "--agent-validation --force"
            ).split(),
            stderr_full=(
                "Warning: Validation result from agent:\n"
                "  pcsmock validation failure\n"
            ),
        )

        self.assert_pcs_success(
            "stonith config".split(),
            dedent(
                """\
                Resource: fencing (class=stonith type=fence_pcsmock_params)
                  Attributes: fencing-instance_attributes
                    ip="is_invalid=True"
                    username=apc
                  Operations:
                    monitor: fencing-monitor-interval-60s
                      interval=60s
                """
            ),
        )

        self.assert_pcs_fail(
            "stonith update fencing bad_ipaddr=test username=login".split(),
            stderr_regexp=(
                "^Error: invalid stonith option 'bad_ipaddr', allowed options"
                " are: [^\n]+, use --force to override\n$"
            ),
        )

        self.assert_pcs_success(
            "stonith update fencing username=testA --agent-validation".split(),
            stderr_start="Warning: The resource was misconfigured before the update,",
        )

        self.assert_pcs_success(
            "stonith config".split(),
            dedent(
                """\
                Resource: fencing (class=stonith type=fence_pcsmock_params)
                  Attributes: fencing-instance_attributes
                    ip="is_invalid=True"
                    username=testA
                  Operations:
                    monitor: fencing-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_create_requires_either_new_or_deprecated(self):
        # 'ipaddr' and 'login' are obsoleted by 'ip' and 'username'
        self.assert_pcs_fail(
            "stonith create test2 fence_pcsmock_params".split(),
            (
                "Error: stonith option 'ip' or 'ipaddr' (deprecated) has to be "
                "specified, use --force to override\n"
                "Error: stonith option 'username' or 'login' (deprecated) has "
                "to be specified, use --force to override\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )

    def test_stonith_create_deprecated_and_obsoleting(self):
        # 'ipaddr' and 'login' are obsoleted by 'ip' and 'username'
        self.assert_pcs_success(
            "stonith create S fence_pcsmock_params ip=i login=l password=1234".split(),
            stderr_full=(
                "Warning: stonith option 'login' is deprecated and might be "
                "removed in a future release, therefore it should not be "
                "used, use 'username' instead\n"
            ),
        )
        self.assert_pcs_success(
            "stonith config S".split(),
            dedent(
                """\
                Resource: S (class=stonith type=fence_pcsmock_params)
                  Attributes: S-instance_attributes
                    ip=i
                    login=l
                    password=1234
                  Operations:
                    monitor: S-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_create_both_deprecated_and_obsoleting(self):
        # 'ipaddr' and 'login' are obsoleted by 'ip' and 'username'
        self.assert_pcs_success(
            (
                "stonith",
                "create",
                "S",
                "fence_pcsmock_params",
                "ip=i1",
                "login=l",
                "ipaddr=i2",
                "username=u",
                "password=1234",
            ),
            stderr_full=(
                "Warning: stonith option 'ipaddr' is deprecated and might be "
                "removed in a future release, therefore it should not be "
                "used, use 'ip' instead\n"
                "Warning: stonith option 'login' is deprecated and might be "
                "removed in a future release, therefore it should not be "
                "used, use 'username' instead\n"
            ),
        )
        self.assert_pcs_success(
            "stonith config S".split(),
            dedent(
                """\
                Resource: S (class=stonith type=fence_pcsmock_params)
                  Attributes: S-instance_attributes
                    ip=i1
                    ipaddr=i2
                    login=l
                    password=1234
                    username=u
                  Operations:
                    monitor: S-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_create_provides_unfencing(self):
        self.assert_pcs_success(
            ("stonith", "create", "f1", "fence_pcsmock_unfencing")
        )
        self.assert_pcs_success(
            (
                "stonith",
                "create",
                "f2",
                "fence_pcsmock_unfencing",
                "meta",
                "provides=unfencing",
            )
        )
        self.assert_pcs_success(
            (
                "stonith",
                "create",
                "f3",
                "fence_pcsmock_unfencing",
                "meta",
                "provides=something",
            )
        )
        self.assert_pcs_success(
            (
                "stonith",
                "create",
                "f4",
                "fence_pcsmock_minimal",
                "meta",
                "provides=something",
            )
        )
        self.assert_pcs_success(
            "stonith config".split(),
            dedent(
                """\
                Resource: f1 (class=stonith type=fence_pcsmock_unfencing)
                  Meta Attributes: f1-meta_attributes
                    provides=unfencing
                  Operations:
                    monitor: f1-monitor-interval-60s
                      interval=60s
                Resource: f2 (class=stonith type=fence_pcsmock_unfencing)
                  Meta Attributes: f2-meta_attributes
                    provides=unfencing
                  Operations:
                    monitor: f2-monitor-interval-60s
                      interval=60s
                Resource: f3 (class=stonith type=fence_pcsmock_unfencing)
                  Meta Attributes: f3-meta_attributes
                    provides=unfencing
                  Operations:
                    monitor: f3-monitor-interval-60s
                      interval=60s
                Resource: f4 (class=stonith type=fence_pcsmock_minimal)
                  Meta Attributes: f4-meta_attributes
                    provides=something
                  Operations:
                    monitor: f4-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_create_action(self):
        self.assert_pcs_fail(
            "stonith create test fence_pcsmock_action action=a".split(),
            (
                "Error: stonith option 'action' is deprecated and might be "
                "removed in a future release, therefore it should not be"
                " used, use 'pcmk_off_action', 'pcmk_reboot_action' instead,"
                " use --force to override\n" + ERRORS_HAVE_OCCURRED
            ),
        )

        self.assert_pcs_success(
            "stonith create test fence_pcsmock_action action=a --force".split(),
            stderr_start=(
                "Warning: stonith option 'action' is deprecated and might be "
                "removed in a future release, therefore it should not be "
                "used, use 'pcmk_off_action', 'pcmk_reboot_action' instead\n"
            ),
        )

        self.assert_pcs_success(
            "stonith config".split(),
            dedent(
                """\
                Resource: test (class=stonith type=fence_pcsmock_action)
                  Attributes: test-instance_attributes
                    action=a
                  Operations:
                    monitor: test-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_create_action_empty(self):
        self.assert_pcs_fail(
            "stonith create test fence_pcsmock_action action=".split(),
            "Error: action cannot be empty\n" + ERRORS_HAVE_OCCURRED,
        )

    def test_stonith_update_action(self):
        self.assert_pcs_success(
            "stonith create test fence_pcsmock_action".split()
        )

        self.assert_pcs_success(
            "stonith config".split(),
            dedent(
                """\
                Resource: test (class=stonith type=fence_pcsmock_action)
                  Operations:
                    monitor: test-monitor-interval-60s
                      interval=60s
                """
            ),
        )

        self.assert_pcs_fail(
            "stonith update test action=a".split(),
            (
                "Error: stonith option 'action' is deprecated and might be "
                "removed in a future release, therefore it should not be"
                " used, use 'pcmk_off_action', 'pcmk_reboot_action' instead,"
                " use --force to override\n"
            ),
        )

        self.assert_pcs_success(
            "stonith update test action=a --force".split(),
            stderr_start=(
                "Warning: stonith option 'action' is deprecated and might be "
                "removed in a future release, therefore it should not be "
                "used, use 'pcmk_off_action', 'pcmk_reboot_action' instead\n"
            ),
        )

        self.assert_pcs_success(
            "stonith config".split(),
            dedent(
                """\
                Resource: test (class=stonith type=fence_pcsmock_action)
                  Attributes: test-instance_attributes
                    action=a
                  Operations:
                    monitor: test-monitor-interval-60s
                      interval=60s
                """
            ),
        )

        self.assert_pcs_success("stonith update test action=".split())

        self.assert_pcs_success(
            "stonith config".split(),
            dedent(
                """\
                Resource: test (class=stonith type=fence_pcsmock_action)
                  Operations:
                    monitor: test-monitor-interval-60s
                      interval=60s
                """
            ),
        )

    def test_stonith_fence_confirm(self):
        self.pcs_runner.cib_file = None
        self.assert_pcs_fail(
            "stonith fence blah blah".split(),
            "Error: must specify one (and only one) node to fence\n",
        )
        self.assert_pcs_fail(
            "stonith confirm blah blah".split(),
            "Error: must specify one (and only one) node to confirm fenced\n",
        )

    def test_no_stonith_warning(self):
        self.assert_pcs_success(
            "stonith create test_stonith fence_pcsmock_minimal".split()
        )

        self.assert_pcs_fail(
            "stonith delete test_stonith".split(),
            (
                "Error: Requested action leaves the cluster with no enabled "
                "means to fence nodes, resulting in the cluster not being able "
                "to recover from certain failure conditions, use --force to "
                "override\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "stonith delete test_stonith --force".split(),
            stderr_full=(
                "Warning: Requested action leaves the cluster with no enabled "
                "means to fence nodes, resulting in the cluster not being able "
                "to recover from certain failure conditions\n"
            ),
        )


_fixture_stonith_level_cache = None
_fixture_stonith_level_cache_lock = Lock()


class StonithLevelTestCibFixture(CachedCibFixture):
    def _fixture_stonith_resource(self, name):
        self.assert_pcs_success(
            [
                "stonith",
                "create",
                name,
                "fence_pcsmock_minimal",
                "pcmk_host_list=rh7-1 rh7-2",
            ]
        )

    def _setup_cib(self):
        self._fixture_stonith_resource("F1")
        self._fixture_stonith_resource("F2")
        self._fixture_stonith_resource("F3")

        self.assert_pcs_success("stonith level add 1 rh7-1 F1".split())
        self.assert_pcs_success("stonith level add 2 rh7-1 F2".split())
        self.assert_pcs_success("stonith level add 2 rh7-2 F1".split())
        self.assert_pcs_success("stonith level add 1 rh7-2 F2".split())
        self.assert_pcs_success("stonith level add 4 regexp%rh7-\\d F3".split())
        self.assert_pcs_success(
            "stonith level add 3 regexp%rh7-\\d F2 F1".split()
        )
        self.assert_pcs_success(
            "stonith level add 5 attrib%fencewith=levels1 F3 F2".split()
        )
        self.assert_pcs_success(
            "stonith level add 6 attrib%fencewith=levels2 F3 F1".split()
        )


CIB_FIXTURE = StonithLevelTestCibFixture(
    "fixture_tier1_stonith_level_tests", rc("cib-empty-withnodes.xml")
)


class LevelTestsBase(TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_test_stonith_level")
        write_file_to_tmpfile(rc("cib-empty-withnodes.xml"), self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.pcs_runner.mock_settings = get_mock_settings()
        self.config = ""
        self.config_lines = []

    def tearDown(self):
        self.temp_cib.close()

    def fixture_stonith_resource(self, name):
        self.assert_pcs_success(
            [
                "stonith",
                "create",
                name,
                "fence_pcsmock_minimal",
                "pcmk_host_list=rh7-1 rh7-2",
            ]
        )

    def fixture_full_configuration(self):
        cib, self.config, self.config_lines = self.fixture_cib_config_cache()
        write_data_to_tmpfile(cib, self.temp_cib)

    def fixture_cib_config_cache(self):
        # pylint: disable=global-statement
        global _fixture_stonith_level_cache  # noqa: PLW0603
        with _fixture_stonith_level_cache_lock:
            if _fixture_stonith_level_cache is None:
                _fixture_stonith_level_cache = self.fixture_cib_config()
            return _fixture_stonith_level_cache

    @staticmethod
    def fixture_cib_config():
        cib_content = ""
        with open(CIB_FIXTURE.cache_path, "r") as cib_file:
            cib_content = cib_file.read()
        config = dedent(
            """\
            Target (node): rh7-1
              Level 1: F1
              Level 2: F2
            Target (node): rh7-2
              Level 1: F2
              Level 2: F1
            Target (regexp): rh7-\\d
              Level 3: F2 F1
              Level 4: F3
            Target (attribute): fencewith=levels1
              Level 5: F3 F2
            Target (attribute): fencewith=levels2
              Level 6: F3 F1
            """
        )
        config_lines = config.splitlines()
        return cib_content, config, config_lines


class LevelBadCommand(LevelTestsBase):
    def test_success(self):
        self.assert_pcs_fail(
            "stonith level nonsense".split(),
            stderr_start="\nUsage: pcs stonith level ...\n",
        )


class LevelAdd(LevelTestsBase):
    def test_not_enough_params(self):
        self.assert_pcs_fail(
            "stonith level add".split(),
            stderr_start="\nUsage: pcs stonith level add...\n",
        )

        self.assert_pcs_fail(
            "stonith level add 1".split(),
            stderr_start="\nUsage: pcs stonith level add...\n",
        )

        self.assert_pcs_fail(
            "stonith level add 1 nodeA".split(),
            stderr_start="\nUsage: pcs stonith level add...\n",
        )

    def test_add_wrong_target_type(self):
        self.assert_pcs_fail(
            "stonith level add 1 error%value F1".split(),
            (
                "Error: 'error' is not an allowed type for 'error%value', "
                "use attrib, node, regexp\n"
            ),
        )

    def test_add_bad_level(self):
        self.fixture_stonith_resource("F1")
        self.assert_pcs_fail(
            "stonith level add NaN rh7-1 F1".split(),
            (
                "Error: 'NaN' is not a valid level value, use 1..9\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_fail(
            "-- stonith level add -10 rh7-1 F1".split(),
            (
                "Error: '-10' is not a valid level value, use 1..9\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_fail(
            "stonith level add 10abc rh7-1 F1".split(),
            (
                "Error: '10abc' is not a valid level value, use 1..9\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_fail(
            "stonith level add 0 rh7-1 F1".split(),
            (
                "Error: '0' is not a valid level value, use 1..9\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_fail(
            "stonith level add 000 rh7-1 F1".split(),
            (
                "Error: '000' is not a valid level value, use 1..9\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )

    def test_add_bad_device(self):
        self.assert_pcs_fail(
            "stonith level add 1 rh7-1 dev@ce".split(),
            (
                "Error: invalid device id 'dev@ce', '@' is not a valid "
                "character for a device id\n" + ERRORS_HAVE_OCCURRED
            ),
        )

    def test_add_more_errors(self):
        self.assert_pcs_fail(
            "stonith level add x rh7-X F0 dev@ce".split(),
            (
                "Error: 'x' is not a valid level value, use 1..9\n"
                "Error: Node 'rh7-X' does not appear to exist in configuration, "
                "use --force to override\n"
                "Error: invalid device id 'dev@ce', '@' is not a valid character "
                "for a device id\n"
                "Error: Stonith resource(s) 'F0' do not exist, use --force to "
                "override\n" + ERRORS_HAVE_OCCURRED
            ),
        )

        self.assert_pcs_fail(
            "stonith level add x rh7-X F0 dev@ce --force".split(),
            (
                "Error: 'x' is not a valid level value, use 1..9\n"
                "Warning: Node 'rh7-X' does not appear to exist in configuration\n"
                "Error: invalid device id 'dev@ce', '@' is not a valid character "
                "for a device id\n"
                "Warning: Stonith resource(s) 'F0' do not exist\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )

    def test_add_level_leading_zero(self):
        self.fixture_stonith_resource("F1")
        self.assert_pcs_success("stonith level add 0002 rh7-1 F1".split())
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (node): rh7-1
                  Level 2: F1
                """
            ),
        )

    def test_add_node(self):
        self.fixture_stonith_resource("F1")
        self.assert_pcs_success("stonith level add 1 rh7-1 F1".split())
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (node): rh7-1
                  Level 1: F1
                """
            ),
        )

        self.assert_pcs_fail(
            "stonith level add 1 rh7-1 F1".split(),
            (
                "Error: Fencing level for 'rh7-1' at level '1' with device(s) "
                "'F1' already exists\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (node): rh7-1
                  Level 1: F1
                """
            ),
        )

    def test_add_node_pattern(self):
        self.fixture_stonith_resource("F1")
        self.assert_pcs_success("stonith level add 1 regexp%rh7-\\d F1".split())
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (regexp): rh7-\\d
                  Level 1: F1
                """
            ),
        )

        self.assert_pcs_fail(
            "stonith level add 1 regexp%rh7-\\d F1".split(),
            (
                r"Error: Fencing level for 'rh7-\d' at level '1' with device(s) "
                "'F1' already exists\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (regexp): rh7-\\d
                  Level 1: F1
                """
            ),
        )

    def test_add_node_attribute(self):
        self.fixture_stonith_resource("F1")
        self.assert_pcs_success(
            "stonith level add 1 attrib%fencewith=levels F1".split()
        )
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (attribute): fencewith=levels
                  Level 1: F1
                """
            ),
        )

        self.assert_pcs_fail(
            "stonith level add 1 attrib%fencewith=levels F1".split(),
            (
                "Error: Fencing level for 'fencewith=levels' at level '1' with "
                "device(s) 'F1' already exists\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (attribute): fencewith=levels
                  Level 1: F1
                """
            ),
        )

    def test_add_more_devices(self):
        self.fixture_stonith_resource("F1")
        self.fixture_stonith_resource("F2")
        self.assert_pcs_success("stonith level add 1 rh7-1 F1 F2".split())
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (node): rh7-1
                  Level 1: F1 F2
                """
            ),
        )

    def test_add_more_devices_old_syntax(self):
        self.fixture_stonith_resource("F1")
        self.fixture_stonith_resource("F2")
        self.fixture_stonith_resource("F3")

        self.assert_pcs_fail(
            "stonith level add 1 rh7-1 F1,F2".split(),
            stderr_full=(
                "Error: invalid device id 'F1,F2', ',' is not a valid character "
                "for a device id\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_fail(
            "stonith level add 2 rh7-1 F1,F2 F3".split(),
            stderr_full=(
                "Error: invalid device id 'F1,F2', ',' is not a valid character "
                "for a device id\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_fail(
            "stonith level add 3 rh7-1 F1 F2,F3".split(),
            stderr_full=(
                "Error: invalid device id 'F2,F3', ',' is not a valid character "
                "for a device id\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "stonith level".split(),
            "",
        )

    def test_nonexistent_node(self):
        self.fixture_stonith_resource("F1")
        self.assert_pcs_fail(
            "stonith level add 1 rh7-X F1".split(),
            (
                "Error: Node 'rh7-X' does not appear to exist in configuration"
                ", use --force to override\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "stonith level add 1 rh7-X F1 --force".split(),
            stderr_full=(
                "Warning: Node 'rh7-X' does not appear to exist in configuration\n"
            ),
        )
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (node): rh7-X
                  Level 1: F1
                """
            ),
        )

    def test_nonexistent_device(self):
        self.assert_pcs_fail(
            "stonith level add 1 rh7-1 F1".split(),
            (
                "Error: Stonith resource(s) 'F1' do not exist"
                ", use --force to override\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "stonith level add 1 rh7-1 F1 --force".split(),
            stderr_full="Warning: Stonith resource(s) 'F1' do not exist\n",
        )
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (node): rh7-1
                  Level 1: F1
                """
            ),
        )

    def test_nonexistent_devices(self):
        self.fixture_stonith_resource("F1")
        self.assert_pcs_fail(
            "stonith level add 1 rh7-1 F1 F2 F3".split(),
            (
                "Error: Stonith resource(s) 'F2', 'F3' do not exist"
                ", use --force to override\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "stonith level add 1 rh7-1 F1 F2 F3 --force".split(),
            stderr_full="Warning: Stonith resource(s) 'F2', 'F3' do not exist\n",
        )
        self.assert_pcs_success(
            "stonith level".split(),
            dedent(
                """\
                Target (node): rh7-1
                  Level 1: F1 F2 F3
                """
            ),
        )


@skip_unless_crm_rule()
class LevelConfig(LevelTestsBase):
    full_config = dedent(
        """\
        Cluster Name: test99
        Corosync Nodes:
         rh7-1 rh7-2
        Pacemaker Nodes:
         rh7-1 rh7-2

        Stonith Devices:{devices}

        Fencing Levels:{levels}
        """
    )

    def test_empty(self):
        self.assert_pcs_success("stonith level config".split(), "")
        self.assert_pcs_success("stonith level".split(), "")
        self.assert_pcs_success(["stonith"], "NO stonith devices configured\n")
        self.pcs_runner.mock_settings["corosync_conf_file"] = rc(
            "corosync.conf"
        )
        self.assert_pcs_success(
            ["config"],
            dedent(
                """\
                Cluster Name: test99
                Corosync Nodes:
                 rh7-1 rh7-2
                Pacemaker Nodes:
                 rh7-1 rh7-2
                """
            ),
        )

    def test_all_possibilities(self):
        self.fixture_full_configuration()
        self.assert_pcs_success("stonith level config".split(), self.config)
        self.assert_pcs_success("stonith level".split(), self.config)
        if PCMK_2_0_3_PLUS:
            result = outdent(
                """\
                  * F1\t(stonith:fence_pcsmock_minimal):\tStopped
                  * F2\t(stonith:fence_pcsmock_minimal):\tStopped
                  * F3\t(stonith:fence_pcsmock_minimal):\tStopped
                """
            )
        else:
            result = outdent(
                """\
                 F1\t(stonith:fence_pcsmock_minimal):\tStopped
                 F2\t(stonith:fence_pcsmock_minimal):\tStopped
                 F3\t(stonith:fence_pcsmock_minimal):\tStopped
                """
            )
        self.assert_pcs_success(
            ["stonith"],
            result
            + "\n"
            + "\n".join(["Fencing Levels:"] + indent(self.config_lines, 1))
            + "\n",
            despace=True,
        )
        self.pcs_runner.mock_settings["corosync_conf_file"] = rc(
            "corosync.conf"
        )
        self.assert_pcs_success(
            ["config"],
            self.full_config.format(
                devices="\n".join(
                    indent(
                        dedent(
                            """
                            Resource: F1 (class=stonith type=fence_pcsmock_minimal)
                              Attributes: F1-instance_attributes
                                pcmk_host_list="rh7-1 rh7-2"
                              Operations:
                                monitor: F1-monitor-interval-60s
                                  interval=60s
                            Resource: F2 (class=stonith type=fence_pcsmock_minimal)
                              Attributes: F2-instance_attributes
                                pcmk_host_list="rh7-1 rh7-2"
                              Operations:
                                monitor: F2-monitor-interval-60s
                                  interval=60s
                            Resource: F3 (class=stonith type=fence_pcsmock_minimal)
                              Attributes: F3-instance_attributes
                                pcmk_host_list="rh7-1 rh7-2"
                              Operations:
                                monitor: F3-monitor-interval-60s
                                  interval=60s
                            """
                        ).splitlines()
                    )
                ),
                levels=("\n" + "\n".join(indent(self.config_lines, 2))),
            ),
        )


class LevelClear(LevelTestsBase):
    def setUp(self):
        super().setUp()
        self.fixture_full_configuration()

    def test_missing_target(self):
        self.assert_pcs_fail(
            "stonith level clear target".split(),
            "Error: Missing value after 'target'\n",
        )

    def test_missing_device(self):
        self.assert_pcs_fail(
            "stonith level clear stonith".split(),
            "Error: Missing value after 'stonith'\n",
        )

    def test_target_and_device(self):
        self.assert_pcs_fail(
            "stonith level clear target rh7-1 stonith F1".split(),
            "Error: Only one of 'target' and 'stonith' can be used\n",
        )

    def test_clear_all(self):
        self.assert_pcs_success("stonith level clear".split())
        self.assert_pcs_success("stonith level config".split(), "")

    def test_clear_nonexistent_node(self):
        self.assert_pcs_fail(
            "stonith level clear target rh-X".split(),
            (
                "Error: Fencing level for 'rh-X' does not exist\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success("stonith level config".split(), self.config)

    def test_clear_nonexistent_devices(self):
        self.assert_pcs_fail(
            "stonith level clear stonith F1 F5".split(),
            (
                "Error: Fencing level with device(s) 'F1', 'F5' does not exist\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success("stonith level config".split(), self.config)

    def test_clear_node(self):
        self.assert_pcs_success("stonith level clear target rh7-1".split())
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[3:]) + "\n",
        )

    def test_clear_pattern(self):
        self.assert_pcs_success(
            "stonith level clear target regexp%rh7-\\d".split()
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:6] + self.config_lines[9:]) + "\n",
        )

    def test_clear_attribute(self):
        self.assert_pcs_success(
            "stonith level clear target attrib%fencewith=levels2".split()
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:11]) + "\n",
        )

    def test_clear_device(self):
        self.assert_pcs_success("stonith level clear stonith F1".split())
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(
                self.config_lines[0:1]
                + self.config_lines[2:5]
                + self.config_lines[6:]
            )
            + "\n",
        )

    def test_clear_devices(self):
        self.assert_pcs_success("stonith level clear stonith F2 F1".split())
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:7] + self.config_lines[8:]) + "\n",
        )

    def test_clear_devices_comma(self):
        # test that old syntax doesn't work in the new command
        self.assert_pcs_fail(
            "stonith level clear stonith F2,F1".split(),
            (
                "Error: invalid stonith id 'F2,F1', ',' is not a valid character "
                "for a stonith id\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success("stonith level config".split(), self.config)

    def _test_missing_stonith_keyword(self):
        self.assert_pcs_fail(
            "stonith level clear target rh7-2 F2".split(),
            "Error: At most one target can be specified\n",
        )

    def _test_missing_target_keyword(self):
        self.assert_pcs_fail(
            "stonith level clear rh7-2 stonith F2".split(),
            "Error: At most one target can be specified\n",
        )

    def _test_missing_target_stonith_keyword(self):
        self.assert_pcs_fail(
            "stonith level clear rh7-2 F2".split(),
            "Error: At most one target can be specified\n",
        )


class LevelDeleteRemove(LevelTestsBase):
    command = None

    def setUp(self):
        super().setUp()
        self.fixture_full_configuration()

    def _test_usage(self):
        self.assert_pcs_fail(
            ["stonith", "level", self.command],
            stderr_start=dedent(
                f"""
                Usage: pcs stonith level {self.command}...
                    level {self.command} <"""
            ),
        )

    def _test_nonexisting_level_node_device(self):
        self.assert_pcs_fail(
            [
                "stonith",
                "level",
                self.command,
                "1",
                "target",
                "rh7-1",
                "stonith",
                "F3",
            ],
            (
                "Error: Fencing level for 'rh7-1' at level '1' with device(s) "
                "'F3' does not exist\n"
            )
            + ERRORS_HAVE_OCCURRED,
        )
        self.assert_pcs_success("stonith level config".split(), self.config)

    def _test_nonexisting_level_pattern_device(self):
        self.assert_pcs_fail(
            [
                "stonith",
                "level",
                self.command,
                "1",
                "target",
                r"regexp%rh7-\d",
                "stonith",
                "F3",
            ],
            (
                "Error: Fencing level for 'rh7-\\d' at level '1' with "
                "device(s) 'F3' does not exist\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success("stonith level config".split(), self.config)

        self.assert_pcs_fail(
            [
                "stonith",
                "level",
                self.command,
                "3",
                "target",
                r"regexp%rh7-\d",
                "stonith",
                "F1",
                "F2",
            ],
            (
                "Error: Fencing level for 'rh7-\\d' at level '3' with "
                "device(s) 'F1', 'F2' does not exist\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success("stonith level config".split(), self.config)

    def _test_nonexisting_level(self):
        self.assert_pcs_fail(
            ["stonith", "level", self.command, "9"],
            (
                "Error: Fencing level at level '9' does not exist\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success("stonith level config".split(), self.config)

    def _test_remove_level(self):
        self.assert_pcs_success(["stonith", "level", self.command, "1"])
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(
                self.config_lines[0:1]
                + self.config_lines[2:4]
                + self.config_lines[5:]
            )
            + "\n",
        )

    def _test_remove_level_node(self):
        self.assert_pcs_success(
            ["stonith", "level", self.command, "1", "target", "rh7-2"]
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:4] + self.config_lines[5:]) + "\n",
        )

    def _test_remove_level_pattern(self):
        self.assert_pcs_success(
            ["stonith", "level", self.command, "3", "target", r"regexp%rh7-\d"]
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:7] + self.config_lines[8:]) + "\n",
        )

    def _test_remove_level_attrib(self):
        self.assert_pcs_success(
            [
                "stonith",
                "level",
                self.command,
                "6",
                "target",
                "attrib%fencewith=levels2",
            ]
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:11]) + "\n",
        )

    def _test_remove_level_device(self):
        self.assert_pcs_success(
            ["stonith", "level", self.command, "1", "stonith", "F2"]
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:4] + self.config_lines[5:]) + "\n",
        )

    def _test_remove_level_devices(self):
        self.assert_pcs_success(
            ["stonith", "level", self.command, "3", "stonith", "F2", "F1"]
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:7] + self.config_lines[8:]) + "\n",
        )

    def _test_comma_separated_devices_not_supported(self):
        self.assert_pcs_fail(
            ["stonith", "level", self.command, "3", "stonith", "F2,F1"],
            "Error: invalid stonith id 'F2,F1', ',' is not a valid character "
            "for a stonith id\n" + ERRORS_HAVE_OCCURRED,
        )

    def _test_remove_level_node_device(self):
        self.assert_pcs_success(
            [
                "stonith",
                "level",
                self.command,
                "1",
                "target",
                "rh7-2",
                "stonith",
                "F2",
            ]
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:4] + self.config_lines[5:]) + "\n",
        )

    def _test_remove_level_device_node(self):
        self.assert_pcs_success(
            [
                "stonith",
                "level",
                self.command,
                "1",
                "stonith",
                "F2",
                "target",
                "rh7-2",
            ]
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:4] + self.config_lines[5:]) + "\n",
        )

    def _test_remove_level_pattern_device(self):
        self.assert_pcs_success(
            [
                "stonith",
                "level",
                self.command,
                "3",
                "target",
                r"regexp%rh7-\d",
                "stonith",
                "F2",
                "F1",
            ]
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:7] + self.config_lines[8:]) + "\n",
        )

    def _test_remove_level_attrib_device(self):
        self.assert_pcs_success(
            [
                "stonith",
                "level",
                self.command,
                "6",
                "target",
                "attrib%fencewith=levels2",
                "stonith",
                "F3",
                "F1",
            ]
        )
        self.assert_pcs_success(
            "stonith level config".split(),
            "\n".join(self.config_lines[:11]) + "\n",
        )

    def _test_missing_stonith_keyword(self):
        self.assert_pcs_fail(
            ["stonith", "level", self.command, "1", "target", "rh7-2", "F2"],
            "Error: At most one target can be specified\n",
        )

    def _test_missing_target_keyword(self):
        self.assert_pcs_fail(
            ["stonith", "level", self.command, "1", "rh7-2", "stonith", "F2"],
            stderr_start=dedent(
                f"""
                Usage: pcs stonith level {self.command}...
                    level {self.command} <"""
            ),
        )

    def _test_missing_target_stonith_keyword(self):
        self.assert_pcs_fail(
            ["stonith", "level", self.command, "1", "rh7-2", "F2"],
            stderr_start=dedent(
                f"""
                Usage: pcs stonith level {self.command}...
                    level {self.command} <"""
            ),
        )

    def _test_missing_target_stonith_values(self):
        self.assert_pcs_fail(
            ["stonith", "level", self.command, "1", "target"],
            "Error: Missing value after 'target'\n",
        )
        self.assert_pcs_fail(
            ["stonith", "level", self.command, "1", "stonith"],
            "Error: Missing value after 'stonith'\n",
        )
        self.assert_pcs_fail(
            ["stonith", "level", self.command, "1", "target", "stonith"],
            "Error: Missing value after 'target' and 'stonith'\n",
        )


class LevelDelete(LevelDeleteRemove, metaclass=ParametrizedTestMetaClass):
    command = "delete"


class LevelRemove(LevelDeleteRemove, metaclass=ParametrizedTestMetaClass):
    command = "remove"


class LevelVerify(LevelTestsBase):
    def test_success(self):
        self.fixture_full_configuration()
        self.assert_pcs_success("stonith level verify".split(), "")

    def test_errors(self):
        self.fixture_stonith_resource("F1")

        self.assert_pcs_success("stonith level add 1 rh7-1 F1".split())
        self.assert_pcs_success(
            "stonith level add 2 rh7-1 FX --force".split(),
            stderr_full="Warning: Stonith resource(s) 'FX' do not exist\n",
        )
        self.assert_pcs_success(
            "stonith level add 1 rh7-X FX --force".split(),
            stderr_full=dedent(
                """\
                Warning: Node 'rh7-X' does not appear to exist in configuration
                Warning: Stonith resource(s) 'FX' do not exist
                """
            ),
        )
        self.assert_pcs_success(
            "stonith level add 2 rh7-Y FY --force".split(),
            stderr_full=dedent(
                """\
                Warning: Node 'rh7-Y' does not appear to exist in configuration
                Warning: Stonith resource(s) 'FY' do not exist
                """
            ),
        )
        self.assert_pcs_success(
            "stonith level add 4 regexp%rh7-\\d FX --force".split(),
            stderr_full="Warning: Stonith resource(s) 'FX' do not exist\n",
        )
        self.assert_pcs_success(
            [
                "stonith",
                "level",
                "add",
                "3",
                r"regexp%rh7-\d",
                "FY",
                "FZ",
                "--force",
            ],
            stderr_full="Warning: Stonith resource(s) 'FY', 'FZ' do not exist\n",
        )

        self.assert_pcs_fail(
            "stonith level verify".split(),
            (
                "Error: Stonith resource(s) 'FX', 'FY', 'FZ' do not exist\n"
                "Error: Node 'rh7-X' does not appear to exist in "
                "configuration\n"
                "Error: Node 'rh7-Y' does not appear to exist in "
                "configuration\n" + ERRORS_HAVE_OCCURRED
            ),
        )


class StonithUpdate(ResourceTest):
    def setUp(self):
        super().setUp()
        self.pcs_runner.mock_settings = get_mock_settings()
        self.fixture_create_stonith()

    def fixture_create_stonith(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_params ip=i login=l ssh=0 debug=d password=1234".split(),
            """
            <resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug" name="debug"
                            value="d"
                        />
                        <nvpair id="S-instance_attributes-ip" name="ip"
                            value="i"
                        />
                        <nvpair id="S-instance_attributes-login" name="login"
                            value="l"
                        />
                        <nvpair id="S-instance_attributes-password" name="password"
                            value="1234"
                        />
                        <nvpair id="S-instance_attributes-ssh" name="ssh"
                            value="0"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
            stderr_full=(
                "Warning: stonith option 'login' is deprecated and might be "
                "removed in a future release, therefore it should not "
                "be used, use 'username' instead\n"
                "Warning: stonith option 'debug' is deprecated and might be "
                "removed in a future release, therefore it should not "
                "be used, use 'debug_file' instead\n"
            ),
        )

    def test_update_meta_attributes(self):
        self.assert_effect(
            "stonith update S meta a=A b=B".split(),
            """
            <resources>
                <primitive id="S" class="stonith" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug" name="debug"
                            value="d"
                        />
                        <nvpair id="S-instance_attributes-ip" name="ip"
                            value="i"
                        />
                        <nvpair id="S-instance_attributes-login" name="login"
                            value="l"
                        />
                        <nvpair id="S-instance_attributes-password"
                            name="password" value="1234"
                        />
                        <nvpair id="S-instance_attributes-ssh" name="ssh"
                            value="0"
                        />
                    </instance_attributes>
                    <operations>
                        <op name="monitor" interval="60s"
                            id="S-monitor-interval-60s"
                        />
                    </operations>
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-a" name="a" value="A"/>
                        <nvpair id="S-meta_attributes-b" name="b" value="B"/>
                    </meta_attributes>
                </primitive>
            </resources>
            """,
            stderr_full=fixture_use_meta_command_instead_warning(
                is_stonith=True
            ),
        )

    def test_set_deprecated_param(self):
        self.assert_effect(
            "stonith update S debug=D".split(),
            """
            <resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug" name="debug"
                            value="D"
                        />
                        <nvpair id="S-instance_attributes-ip" name="ip"
                            value="i"
                        />
                        <nvpair id="S-instance_attributes-login" name="login"
                            value="l"
                        />
                        <nvpair id="S-instance_attributes-password" name="password"
                            value="1234"
                        />
                        <nvpair id="S-instance_attributes-ssh" name="ssh"
                            value="0"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
            stderr_full=(
                "Warning: stonith option 'debug' is deprecated and might be "
                "removed in a future release, therefore it should not "
                "be used, use 'debug_file' instead\n"
            ),
        )

    def test_unset_deprecated_param(self):
        self.assert_effect(
            "stonith update S debug=".split(),
            """
            <resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-ip" name="ip"
                            value="i"
                        />
                        <nvpair id="S-instance_attributes-login" name="login"
                            value="l"
                        />
                        <nvpair id="S-instance_attributes-password" name="password"
                            value="1234"
                        />
                        <nvpair id="S-instance_attributes-ssh" name="ssh"
                            value="0"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_unset_deprecated_required_param(self):
        self.assert_pcs_fail(
            "stonith update S login=".split(),
            "Error: stonith option 'username' or 'login' (deprecated) has to "
            "be specified, use --force to override\n",
        )

    def test_set_obsoleting_param(self):
        self.assert_effect(
            "stonith update S ssh=1".split(),
            """
            <resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug" name="debug"
                            value="d"
                        />
                        <nvpair id="S-instance_attributes-ip" name="ip"
                            value="i"
                        />
                        <nvpair id="S-instance_attributes-login" name="login"
                            value="l"
                        />
                        <nvpair id="S-instance_attributes-password" name="password"
                            value="1234"
                        />
                        <nvpair id="S-instance_attributes-ssh" name="ssh"
                            value="1"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_unset_obsoleting_param(self):
        self.assert_effect(
            "stonith update S ssh=".split(),
            """
            <resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug" name="debug"
                            value="d"
                        />
                        <nvpair id="S-instance_attributes-ip" name="ip"
                            value="i"
                        />
                        <nvpair id="S-instance_attributes-login" name="login"
                            value="l"
                        />
                        <nvpair id="S-instance_attributes-password" name="password"
                            value="1234"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_unset_obsoleting_required_param(self):
        self.assert_pcs_fail(
            "stonith update S ip=".split(),
            "Error: stonith option 'ip' or 'ipaddr' (deprecated) has to be "
            "specified, use --force to override\n",
        )

    def test_unset_deprecated_required_set_obsoleting(self):
        self.assert_effect(
            "stonith update S login= username=u".split(),
            """
            <resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug" name="debug"
                            value="d"
                        />
                        <nvpair id="S-instance_attributes-ip" name="ip"
                            value="i"
                        />
                        <nvpair id="S-instance_attributes-password" name="password"
                            value="1234"
                        />
                        <nvpair id="S-instance_attributes-ssh" name="ssh"
                            value="0"
                        />
                        <nvpair id="S-instance_attributes-username"
                            name="username" value="u"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_unset_obsoleting_required_set_deprecated(self):
        self.assert_effect(
            "stonith update S ip= ipaddr=I".split(),
            """
            <resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug" name="debug"
                            value="d"
                        />
                        <nvpair id="S-instance_attributes-login" name="login"
                            value="l"
                        />
                        <nvpair id="S-instance_attributes-password" name="password"
                            value="1234"
                        />
                        <nvpair id="S-instance_attributes-ssh" name="ssh"
                            value="0"
                        />
                        <nvpair id="S-instance_attributes-ipaddr" name="ipaddr"
                            value="I"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
            stderr_full=(
                "Warning: stonith option 'ipaddr' is deprecated and might be "
                "removed in a future release, therefore it should not "
                "be used, use 'ip' instead\n"
            ),
        )

    def test_set_both_deprecated_and_obsoleting(self):
        self.assert_effect(
            "stonith update S ip=I1 ipaddr=I2".split(),
            """
            <resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug" name="debug"
                            value="d"
                        />
                        <nvpair id="S-instance_attributes-ip" name="ip"
                            value="I1"
                        />
                        <nvpair id="S-instance_attributes-login" name="login"
                            value="l"
                        />
                        <nvpair id="S-instance_attributes-password" name="password"
                            value="1234"
                        />
                        <nvpair id="S-instance_attributes-ssh" name="ssh"
                            value="0"
                        />
                        <nvpair id="S-instance_attributes-ipaddr" name="ipaddr"
                            value="I2"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
            stderr_full=(
                "Warning: stonith option 'ipaddr' is deprecated and might be "
                "removed in a future release, therefore it should not "
                "be used, use 'ip' instead\n"
            ),
        )



from collections import defaultdict
from dataclasses import dataclass
from itertools import chain
from typing import Iterable, Mapping, cast

from lxml import etree
from lxml.etree import _Element

from pcs.common import reports
from pcs.common.resource_status import (
    MoreChildrenQuantifierType,
    ResourcesStatusFacade,
    ResourceState,
)
from pcs.common.types import (
    StringCollection,
    StringSequence,
)
from pcs.lib.cib import const
from pcs.lib.cib.constraint.common import (
    is_constraint,
    is_set_constraint,
)
from pcs.lib.cib.constraint.location import (
    is_location_constraint,
    is_location_rule,
)
from pcs.lib.cib.fencing_topology import (
    find_levels_with_device,
    has_any_devices,
    remove_device_from_level,
)
from pcs.lib.cib.resource.clone import is_any_clone
from pcs.lib.cib.resource.common import get_inner_resources, is_resource
from pcs.lib.cib.resource.group import is_group
from pcs.lib.cib.tag import is_tag
from pcs.lib.cib.tools import (
    ElementNotFound,
    find_elements_referencing_id,
    get_element_by_id,
    get_elements_by_ids,
    get_fencing_topology,
    remove_element_by_id,
    remove_one_element,
)
from pcs.lib.pacemaker.live import parse_cib_xml
from pcs.lib.pacemaker.state import ensure_resource_state
from pcs.lib.pacemaker.status import (
    ClusterStatusParser,
    ClusterStatusParsingError,
    cluster_status_parsing_error_to_report,
)
from pcs.lib.xml_tools import get_root


@dataclass(frozen=True)
class DependantElements:
    id_tag_map: dict[str, str]

    def to_reports(self) -> reports.ReportItemList:
        if not self.id_tag_map:
            return []

        return [
            reports.ReportItem.info(
                reports.messages.CibRemoveDependantElements(self.id_tag_map)
            )
        ]


@dataclass(frozen=True)
class ElementReferences:
    reference_map: dict[str, set[str]]
    id_tag_map: dict[str, str]

    def to_reports(self) -> reports.ReportItemList:
        if not self.reference_map:
            return []

        return [
            reports.ReportItem.info(
                reports.messages.CibRemoveReferences(
                    self.id_tag_map, self.reference_map
                )
            )
        ]


@dataclass(frozen=True)
class UnsupportedElements:
    id_tag_map: Mapping[str, str]
    supported_element_types: StringCollection


class ElementsToRemove:
    """
    Find ids of all elements that should be removed. This function is aware
    of relations and references between elements and will also return ids of
    elements that are somehow referencing elements with specified ids.

    cib -- the whole cib
    ids -- ids of configuration elements to remove
    """

    def __init__(self, cib: _Element, ids: StringCollection):
        wip_cib = parse_cib_xml(etree.tostring(cib).decode())

        initial_ids = set(ids)
        elements_to_process, missing_ids = get_elements_by_ids(
            wip_cib, initial_ids
        )

        supported_elements, unsupported_elements = _validate_element_types(
            elements_to_process
        )

        element_ids_to_remove, removing_references_from = (
            _get_dependencies_to_remove(supported_elements)
        )

        # We need to use ids of the elements, since we will work with cib, but
        # the the elements were instantiated using the wip_cib, which means we
        # cannot reuse the elements
        self._ids_to_remove = element_ids_to_remove
        self._dependant_element_ids = self._ids_to_remove - initial_ids
        self._missing_ids = set(missing_ids)
        self._unsupported_ids = {
            str(el.attrib["id"]) for el in unsupported_elements
        }

        all_ids = set(
            chain(
                self._ids_to_remove,
                self._unsupported_ids,
                *removing_references_from.values(),
            )
        )
        self._id_tag_map = {
            str(el.attrib["id"]): el.tag
            for el in get_elements_by_ids(cib, all_ids)[0]
        }
        self._element_references = removing_references_from
        self._resources_to_remove = [
            el
            for el in get_elements_by_ids(cib, sorted(element_ids_to_remove))[0]
            if is_resource(el)
        ]

    @property
    def ids_to_remove(self) -> set[str]:
        """
        Ids of ALL cib elements with id that should be removed, including all
        resource and dependant elements
        """
        return set(self._ids_to_remove)

    @property
    def resources_to_remove(self) -> list[_Element]:
        """
        List of cib resources that should be removed. Used for operations
        needed for resources before their deletion, e.g. disabling them.
        """
        return list(self._resources_to_remove)

    @property
    def dependant_elements(self) -> DependantElements:
        """
        Information about cib elements that are removed indirectly as
        dependencies of other removed cib elements
        """
        return DependantElements(
            {
                element_id: self._id_tag_map[element_id]
                for element_id in self._dependant_element_ids
            }
        )

    @property
    def element_references(self) -> ElementReferences:
        """
        Information about cib element references that need to be removed.
        These references does not have their own id and need special handling
        while removing, e.g. references to stonith devices in fencing-level
        elements, or obj-ref elements inside tag elements.
        """
        return ElementReferences(
            dict(self._element_references),
            {
                element_id: self._id_tag_map[element_id]
                for element_id in chain(
                    self._element_references,
                    *self._element_references.values(),
                )
            },
        )

    @property
    def missing_ids(self) -> set[str]:
        """
        Set of ids not present in the cib
        """
        return set(self._missing_ids)

    @property
    def unsupported_elements(self) -> UnsupportedElements:
        """
        Information about cib elements that cannot be removed using this
        mechanism
        """
        return UnsupportedElements(
            id_tag_map={
                element_id: self._id_tag_map[element_id]
                for element_id in self._unsupported_ids
            },
            # the list of tags should match the validations done in
            # _validate_element_types function
            supported_element_types=["constraint", "location rule", "resource"],
        )


def ensure_resources_stopped(
    state: _Element,
    resource_ids: StringSequence,
    force_flags: reports.types.ForceFlags,
) -> reports.ReportItemList:
    """
    Ensure that all resources that should be stopped are stopped.

    state -- state of the cluster
    elements -- elements planned to be removed
    """
    not_stopped_ids = []
    report_list: reports.ReportItemList = []
    try:
        parser = ClusterStatusParser(state)
        try:
            status_dto = parser.status_xml_to_dto()
        except ClusterStatusParsingError as e:
            report_list.append(cluster_status_parsing_error_to_report(e))
            return report_list
        report_list.extend(parser.get_warnings())

        status = ResourcesStatusFacade.from_resources_status_dto(status_dto)
        for r_id in resource_ids:
            if not status.exists(r_id, None):
                # Pacemaker does not put misconfigured resources into cluster
                # status and we are unable to check state of such resources.
                # This happens for e.g. undle with primitive resource inside and
                # no IP address for the bundle specified. We expect the resource
                # to be stopped since it is misconfigured.
                report_list.append(
                    reports.ReportItem.debug(
                        reports.messages.ConfiguredResourceMissingInStatus(
                            r_id, ResourceState.STOPPED
                        )
                    )
                )
            elif not status.is_state(
                r_id,
                None,
                ResourceState.STOPPED,
                instances_quantifier=(
                    MoreChildrenQuantifierType.ALL
                    if status.can_have_multiple_instances(r_id)
                    else None
                ),
            ):
                not_stopped_ids.append(r_id)
    except NotImplementedError:
        # TODO remove when issue with bundles in status is fixed
        not_stopped_ids = [
            resource_id
            for resource_id in resource_ids
            if ensure_resource_state(False, state, resource_id).severity.level
            == reports.item.ReportItemSeverity.ERROR
        ]

    if not_stopped_ids:
        report_list.append(
            reports.ReportItem(
                reports.get_severity(
                    reports.codes.FORCE, reports.codes.FORCE in force_flags
                ),
                reports.messages.CannotRemoveResourcesNotStopped(
                    not_stopped_ids
                ),
            )
        )

    return report_list


def remove_specified_elements(
    cib: _Element, elements: ElementsToRemove
) -> None:
    """
    Remove all elements that need to be removed.

    state -- state of the cluster
    elements -- elements planned to be removed
    """
    for element_id in elements.ids_to_remove:
        remove_element_by_id(cib, element_id)

    element_references = elements.element_references
    for (
        referenced_id,
        referenced_in_ids,
    ) in element_references.reference_map.items():
        for element_id in referenced_in_ids:
            _remove_element_reference(
                cib,
                referenced_id,
                element_id,
                element_references.id_tag_map[element_id],
            )


def _validate_element_types(
    elements: Iterable[_Element],
) -> tuple[list[_Element], list[_Element]]:
    supported_elements = []
    unsupported_elements = []

    for el in elements:
        # valid elements should match the valid tags reported from
        # ElementsToRemove.unsupported_elements property
        if is_constraint(el) or is_location_rule(el) or is_resource(el):
            supported_elements.append(el)
        else:
            unsupported_elements.append(el)

    return supported_elements, unsupported_elements


_REFERENCE_TAG_XPATH_MAP = {
    const.TAG_RESOURCE_SET: f"./{const.TAG_RESOURCE_REF}[@id=$referenced_id]",
    const.TAG_TAG: f"./{const.TAG_OBJREF}[@id=$referenced_id]",
}


def _remove_element_reference(
    cib: _Element,
    element_id: str,
    referenced_in_id: str,
    referenced_in_tag: str,
) -> None:
    # If element has id, then it was already removed using its id and does not
    # need to be removed using this reference mapping. Therefore, we need to
    # only remove elements that do not have id here, such as obj_ref and
    # resource_ref.
    try:
        element = get_element_by_id(cib, referenced_in_id)
    except ElementNotFound:
        return

    if referenced_in_tag == const.TAG_FENCING_LEVEL:
        remove_device_from_level(element, element_id)
        return

    if referenced_in_tag not in _REFERENCE_TAG_XPATH_MAP:
        return
    for el in cast(
        list[_Element],
        element.xpath(
            _REFERENCE_TAG_XPATH_MAP[referenced_in_tag],
            referenced_id=element_id,
        ),
    ):
        remove_one_element(el)


def _get_dependencies_to_remove(
    elements: Iterable[_Element],
) -> tuple[set[str], dict[str, set[str]]]:
    """
    Get ids of all elements that need to be removed (including specified
    elements) together with specified elements based on their relations.
    Also return mapping for elements whose references are going to be
    deleted from their respective parent elements, without deleting the
    parent itself.

    WARNING: this is a destructive operation for elements and their etree.

    elements -- iterable of elements that are planned to be removed
    """
    elements_to_process = list(elements)
    element_ids_to_remove: set[str] = set()
    removing_references_from: dict[str, set[str]] = defaultdict(set)

    while elements_to_process:
        el = elements_to_process.pop(0)
        element_id = str(el.attrib["id"])

        # Elements with these tags are only used for referencing other elements.
        # The 'id' attribute in these does not represent the id of the element
        # itself, but the id of the element that they refer to.
        # Therefore, it does not make sense to try finding any references to
        # these elements.
        if el.tag not in (
            const.TAG_OBJREF,
            const.TAG_RESOURCE_REF,
            const.TAG_ROLE,
        ):
            if element_id in element_ids_to_remove:
                continue
            element_ids_to_remove.add(element_id)
            elements_to_process.extend(_get_element_references(el))
            elements_to_process.extend(_get_inner_references(el))

            for level_el in find_levels_with_device(
                get_fencing_topology(get_root(el)), element_id
            ):
                removing_references_from[element_id].add(
                    str(level_el.attrib["id"])
                )
                remove_device_from_level(level_el, element_id)
                if not has_any_devices(level_el):
                    elements_to_process.append(level_el)

        parent_el = el.getparent()
        if parent_el is not None:
            # We only want to remove parent elements that are invalid when
            # empty. There may be ACLs set in pacemaker which allow "write" for
            # the child elements (adding, changing and removing) but not their
            # parent elements. In such case, removing the parent element would
            # cause the whole change to be rejected by pacemaker with a
            # "permission denied" message.
            # https://bugzilla.redhat.com/show_bug.cgi?id=1642514
            if _is_empty_after_inner_el_removal(parent_el):
                elements_to_process.append(parent_el)
            parent_el.remove(el)

            parent_id = parent_el.get("id")
            if parent_id is not None:
                removing_references_from[element_id].add(parent_id)

    # Removing references from parent elements that are going to be removed
    # (they are present in the 'element_ids_to_remove') is unncesary, since all
    # of the child elements are removed when parent is removed. This means we
    # can filter out such references from the resulting mapping.
    for key in list(removing_references_from):
        removing_references_from[key].difference_update(element_ids_to_remove)
        if not removing_references_from[key]:
            del removing_references_from[key]

    return element_ids_to_remove, removing_references_from


def _get_element_references(element: _Element) -> Iterable[_Element]:
    """
    Return all CIB elements that are referencing specified element

    element -- references to this element will be
    """
    return find_elements_referencing_id(element, str(element.attrib["id"]))


def _get_inner_references(element: _Element) -> Iterable[_Element]:
    """
    Get all inner elements with attribute id, which means that they might be
    referenced in IDREF. Elements with attribute id and type IDREF are also
    returned.
    """
    # return cast(Iterable[_Element], element.xpath("./*[@id]"))
    if is_resource(element):
        try:
            # we are removing elements from the tree, therefore assertions of
            # this function can fail
            return get_inner_resources(element)
        except IndexError:
            return []
    # if element.tag == "alert":
    #     return element.findall("recipient")
    # if is_set_constraint(element):
    #     return element.findall("resource_set")
    # if element.tag == "acl_role":
    #     return element.findall("acl_permission")
    return []


def _is_last_element(parent_element: _Element, child_tag: str) -> bool:
    return len(parent_element.findall(f"./{child_tag}")) == 1


def _is_empty_after_inner_el_removal(  # noqa: PLR0911
    parent_el: _Element,
) -> bool:
    # pylint: disable=too-many-return-statements
    if is_any_clone(parent_el):
        return True
    if is_group(parent_el):
        return len(get_inner_resources(parent_el)) == 1
    if is_tag(parent_el):
        return _is_last_element(parent_el, const.TAG_OBJREF)
    if parent_el.tag == const.TAG_RESOURCE_SET:
        return _is_last_element(parent_el, const.TAG_RESOURCE_REF)
    if is_set_constraint(parent_el):
        return _is_last_element(parent_el, const.TAG_RESOURCE_SET)
    if is_location_constraint(parent_el):
        return _is_last_element(parent_el, const.TAG_RULE)
    return False



from unittest import TestCase

from lxml import etree

from pcs.common import reports
from pcs.lib.cib import element_description

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal
from pcs_test.tools.xml import etree_to_str, str_to_etree


class ValidateDescriptionSupport(TestCase):
    def test_supported_element_no_reports(self):
        supported_elements = [
            '<primitive id="A" />',
            '<bundle id="A" />',
            '<clone id="A" />',
            '<group id="A" />',
            '<acl_permission id="A" />',
            '<alert id="A" />',
            '<node id="A" />',
            '<recipient id="A" />',
        ]

        for element_str in supported_elements:
            element = str_to_etree(element_str)
            with self.subTest(element=element_str):
                assert_report_item_list_equal(
                    element_description.validate_description_support(element),
                    [],
                )

    def test_unusported_element_report(self):
        supported_elements = [
            '<tag id="A" />',
            '<rsc_colocation id="A" />',
            '<rsc_resource_set id="A" />',
            '<rule id="A" />',
        ]

        for element_str in supported_elements:
            element = str_to_etree(element_str)
            with self.subTest(element=element_str):
                assert_report_item_list_equal(
                    element_description.validate_description_support(element),
                    [
                        fixture.error(
                            reports.codes.ID_DOES_NOT_SUPPORT_ELEMENT_DESCRIPTIONS,
                            element_id="A",
                            expected_types=sorted(
                                element_description.TAG_LIST_SUPPORTS_DESCRIPTION
                            ),
                            element_type=element.tag,
                        )
                    ],
                )


class SetDescription(TestCase):
    def test_add_description(self):
        element = etree.fromstring('<primitive id="A"/>')

        element_description.set_description(element, "X")

        self.assertEqual(
            etree_to_str(element), '<primitive id="A" description="X"/>\n'
        )

    def test_update_description(self):
        element = etree.fromstring('<primitive id="A" description="X"/>')

        element_description.set_description(element, "Y")

        self.assertEqual(
            etree_to_str(element), '<primitive id="A" description="Y"/>\n'
        )

    def test_remove_description(self):
        element = etree.fromstring('<primitive id="A" description="X"/>')

        element_description.set_description(element, "")

        self.assertEqual(etree_to_str(element), '<primitive id="A"/>\n')


class GetDescription(TestCase):
    def test_read_description(self):
        element = etree.fromstring('<primitive id="A" description="X"/>')

        result = element_description.get_description(element)

        self.assertEqual(result, "X")

    def test_no_description(self):
        element = etree.fromstring('<primitive id="A"/>')

        result = element_description.get_description(element)

        self.assertEqual(result, "")



from textwrap import dedent
from unittest import TestCase

from lxml import etree

from pcs.common import const
from pcs.common.str_tools import format_list

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import XmlManipulation

ERRORS_HAVE_OCCURRED = (
    "Error: Errors have occurred, therefore pcs is unable to continue\n"
)
RULE_ARGV_DEPRECATED = (
    "Deprecation Warning: Specifying a rule as multiple arguments is "
    "deprecated and might be removed in a future release, specify the rule as "
    "a single string instead\n"
)
empty_cib = get_test_resource("cib-empty-3.9.xml")


class RuleBaseMixin(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(etree.parse(cib).find(".//constraints"))
    ),
    AssertPcsMixin,
):
    xml_resource = """
        <primitive id="R1" class="ocf" provider="pacemaker" type="Dummy" />
    """
    xml_constraint_simple = """
        <rsc_location id="loc-simple" rsc="R1" node="node1" score="7" />
    """
    xml_constraint_rule = """
        <rsc_location id="loc-rule" rsc="R1">
          <rule id="loc-rule-rule" boolean-op="and" score="INFINITY">
            <expression id="loc-rule-rule-expr"
                attribute="#uname" operation="eq" value="node2"
            />
          </rule>
          {second_rule}
        </rsc_location>
    """
    xml_constraint_rule_duplicate = """
        <rsc_location id="loc-rule-dup" rsc="R1">
          <rule id="loc-rule-dup-rule" boolean-op="and" score="INFINITY">
            <expression id="loc-rule-dup-rule-expr"
                attribute="#uname" operation="eq" value="node3"
            />
          </rule>
        </rsc_location>
    """

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_constraint_location_rule")
        self.pcs_runner = PcsRunner(self.temp_cib.name)

        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("resources", self.xml_resource)
        xml_manip.append_to_first_tag_name(
            "constraints",
            self.xml_constraint_simple,
            self.xml_constraint_rule.format(second_rule=""),
            self.xml_constraint_rule_duplicate,
        )
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

    def tearDown(self):
        self.temp_cib.close()


class CreateWithRule(RuleBaseMixin, TestCase):
    def fixture_constraints(self, additional_xml):
        return (
            "<constraints>"
            + self.xml_constraint_simple
            + self.xml_constraint_rule.format(second_rule="")
            + self.xml_constraint_rule_duplicate
            + additional_xml
            + "</constraints>"
        )

    def test_success_minimal(self):
        self.assert_effect(
            [
                "constraint location R1 rule".split() + ["#uname eq node1"],
                "constraint location %R1 rule".split() + ["#uname eq node1"],
                (
                    "constraint location resource%R1 rule".split()
                    + ["#uname eq node1"]
                ),
            ],
            self.fixture_constraints(
                """
                <rsc_location id="location-R1" rsc="R1">
                  <rule id="location-R1-rule" boolean-op="and" score="INFINITY">
                    <expression id="location-R1-rule-expr"
                        attribute="#uname" operation="eq" value="node1" />
                  </rule>
                </rsc_location>
                """
            ),
        )

    def test_success_minimal_deprecated_form(self):
        self.assert_effect(
            [
                "constraint location R1 rule #uname eq node1".split(),
                "constraint location %R1 rule #uname eq node1".split(),
                "constraint location resource%R1 rule #uname eq node1".split(),
            ],
            self.fixture_constraints(
                """
                <rsc_location id="location-R1" rsc="R1">
                  <rule id="location-R1-rule" boolean-op="and" score="INFINITY">
                    <expression id="location-R1-rule-expr"
                        attribute="#uname" operation="eq" value="node1" />
                  </rule>
                </rsc_location>
                """
            ),
            stderr_full=RULE_ARGV_DEPRECATED,
        )

    def test_success_all_options(self):
        self.assert_effect(
            (
                "constraint location regexp%R\\d+ rule id=my-rule score=7 "
                "constraint-id=my-loc resource-discovery=always role=Promoted "
            ).split()
            + ["#uname eq node1"],
            self.fixture_constraints(
                """
                <rsc_location id="my-loc" rsc-pattern="R\\d+"
                    resource-discovery="always"
                >
                  <rule id="my-rule" boolean-op="and" score="7" role="Promoted">
                    <expression id="my-rule-expr"
                        attribute="#uname" operation="eq" value="node1" />
                  </rule>
                </rsc_location>
                """
            ),
        )

    def test_duplicate_constraint(self):
        self.assert_pcs_fail(
            "constraint location R1 rule".split() + ["#uname eq node2"],
            stderr_full=dedent(
                """\
                Duplicate constraints:
                  resource 'R1' (id: loc-rule)
                    Rules:
                      Rule: boolean-op=and score=INFINITY (id: loc-rule-rule)
                        Expression: #uname eq node2 (id: loc-rule-rule-expr)
                Error: Duplicate constraint already exists, use --force to override
                """
                + ERRORS_HAVE_OCCURRED
            ),
        )

    def test_duplicate_constraint_forced(self):
        self.assert_effect(
            "--force constraint location R1 rule".split() + ["#uname eq node2"],
            self.fixture_constraints(
                """
                <rsc_location id="location-R1" rsc="R1">
                  <rule id="location-R1-rule" boolean-op="and" score="INFINITY">
                    <expression id="location-R1-rule-expr"
                        attribute="#uname" operation="eq" value="node2" />
                  </rule>
                </rsc_location>
                """
            ),
            stderr_full=dedent(
                """\
                Duplicate constraints:
                  resource 'R1' (id: loc-rule)
                    Rules:
                      Rule: boolean-op=and score=INFINITY (id: loc-rule-rule)
                        Expression: #uname eq node2 (id: loc-rule-rule-expr)
                Warning: Duplicate constraint already exists
                """
            ),
        )

    def test_errors(self):
        roles = format_list(
            [const.PCMK_ROLE_PROMOTED, const.PCMK_ROLE_UNPROMOTED]
        )
        self.assert_pcs_fail(
            (
                "constraint location R1 rule resource-discovery=badly "
                "role=bad-role bad=option #uname eq"
            ).split(),
            stderr_full=RULE_ARGV_DEPRECATED
            + dedent(
                f"""\
                Error: 'badly' is not a valid resource-discovery value, use 'always', 'exclusive', 'never', use --force to override
                Error: 'bad-role' is not a valid role value, use {roles}
                Error: 'bad=option #uname eq' is not a valid rule expression, parse error near or after line 1 column 12
                  bad=option #uname eq
                  -----------^
                """
                + ERRORS_HAVE_OCCURRED
            ),
        )



from unittest import (
    TestCase,
    mock,
)

from pcs.cli.cluster import command
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import InputModifiers
from pcs.common import reports
from pcs.common.reports import codes as report_codes
from pcs.lib.errors import LibraryError

from pcs_test.tools.misc import dict_to_modifiers

UNSTOPPED_RESOURCES_ERROR_REPORT = reports.ReportItem.error(
    reports.messages.CannotRemoveResourcesNotStopped(["A"])
)
NODE_NOT_FOUND_ERROR_REPORT = reports.ReportItem.error(
    reports.messages.NodeNotFound("A", ["remote"])
)
FORCEABLE_ERROR_REPORT = reports.ReportItem.error(
    reports.messages.NodeNotFound("A", ["remote"]),
    force_code=reports.codes.FORCE,
)
INFO_REPORT = reports.ReportItem.info(
    reports.messages.CibRemoveDependantElements({})
)


class ParseNodeAddRemote(TestCase):
    # pylint: disable=protected-access
    def test_deal_with_explicit_address(self):
        self.assertEqual(
            command._node_add_remote_separate_name_and_addr(
                ["name", "address", "a=b"]
            ),
            ("name", "address", ["a=b"]),
        )

    def test_deal_with_implicit_address(self):
        self.assertEqual(
            command._node_add_remote_separate_name_and_addr(["name", "a=b"]),
            ("name", None, ["a=b"]),
        )


class NodeRemoveRemoteBase:
    def setUp(self):
        self.lib = mock.Mock(
            spec_set=["cluster", "env", "remote_node", "resource"]
        )
        self.lib.env = mock.Mock(spec_set=["report_processor"])
        self.lib.remote_node = mock.Mock(
            spec_set=["get_resource_ids", "node_remove_remote"]
        )
        self.lib.cluster = mock.Mock(spec_set=["wait_for_pcmk_idle"])
        self.lib.resource = mock.Mock(spec_set=["stop"])

        self.report_processor_patcher = mock.patch(
            "pcs.cli.cluster.command.NodeRemoveRemoteReportProcessor",
            new_callable=mock.Mock,
            spec_set=["reports", "already_reported_to_console"],
        )
        self.report_processor = self.report_processor_patcher.start()
        self.report_processor.reports = mock.PropertyMock(return_value=[])
        self.report_processor.already_reported_to_console = mock.PropertyMock(
            return_value=False
        )
        # https://docs.python.org/3/library/unittest.mock.html#unittest.mock.PropertyMock
        # We need to attach PropertyMocks to the mock type objects
        type(
            self.report_processor.return_value
        ).reports = self.report_processor.reports
        type(
            self.report_processor.return_value
        ).already_reported_to_console = (
            self.report_processor.already_reported_to_console
        )

    def tearDown(self):
        self.report_processor_patcher.stop()

    def _call_cmd(self, argv, modifiers=None):
        command.node_remove_remote(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def assert_lib_calls(self, expected_calls):
        self.lib.assert_has_calls(expected_calls)
        self.assertEqual(len(self.lib.mock_calls), len(expected_calls))

    def test_no_args(self, mock_process_library_reports):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.assert_lib_calls([])
        self.report_processor.assert_not_called()
        self.report_processor.reports.assert_not_called()
        self.report_processor.already_reported_to_console.assert_not_called()
        mock_process_library_reports.assert_not_called()

    def test_too_many_args(self, mock_process_library_reports):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["A", "B"])
        self.assertIsNone(cm.exception.message)
        self.assert_lib_calls([])
        self.report_processor.assert_not_called()
        self.report_processor.reports.assert_not_called()
        self.report_processor.already_reported_to_console.assert_not_called()
        mock_process_library_reports.assert_not_called()

    def test_success_should_print_reports(self, mock_process_library_reports):
        self.report_processor.reports.return_value = [INFO_REPORT]
        self.report_processor.already_reported_to_console.return_value = False
        self._call_cmd(["A"])

        self.report_processor.assert_called_once_with(False)
        self.assert_lib_calls(
            [mock.call.remote_node.node_remove_remote("A", [])]
        )
        mock_process_library_reports.assert_called_once_with(
            [INFO_REPORT], include_debug=False
        )

    def test_success_reports_already_reported(
        self, mock_process_library_reports
    ):
        self.report_processor.already_reported_to_console.return_value = True
        self._call_cmd(["A"])

        self.report_processor.assert_called_once_with(False)
        self.assert_lib_calls(
            [mock.call.remote_node.node_remove_remote("A", [])]
        )
        mock_process_library_reports.assert_not_called()

    def test_success_with_debug(self, mock_process_library_reports):
        self.report_processor.reports.return_value = [INFO_REPORT]
        self.report_processor.already_reported_to_console.return_value = False
        self._call_cmd(["A"], {"debug": True})

        self.report_processor.assert_called_once_with(True)
        self.assert_lib_calls(
            [mock.call.remote_node.node_remove_remote("A", [])]
        )
        mock_process_library_reports.assert_called_once_with(
            [INFO_REPORT], include_debug=True
        )

    def test_skip_offline(self, mock_process_library_reports):
        self.report_processor.reports.return_value = [INFO_REPORT]
        self.report_processor.already_reported_to_console.return_value = False
        self._call_cmd(["A"], {"skip-offline": True})

        self.report_processor.assert_called_once_with(False)
        self.assert_lib_calls(
            [
                mock.call.remote_node.node_remove_remote(
                    "A", [reports.codes.SKIP_OFFLINE_NODES]
                )
            ]
        )
        mock_process_library_reports.assert_called_once_with(
            [INFO_REPORT], include_debug=False
        )

    def test_dont_stop_me_now(self, mock_process_library_reports):
        self._call_cmd(["A"], {"no-stop": True})

        self.report_processor.assert_not_called()
        self.assert_lib_calls(
            [mock.call.remote_node.node_remove_remote("A", [])]
        )
        mock_process_library_reports.assert_not_called()

    def test_no_stop_all_force_flags(self, mock_process_library_reports):
        self._call_cmd(
            ["A"], {"force": True, "no-stop": True, "skip-offline": True}
        )

        self.report_processor.assert_not_called()
        self.assert_lib_calls(
            [
                mock.call.remote_node.node_remove_remote(
                    "A", [reports.codes.FORCE, reports.codes.SKIP_OFFLINE_NODES]
                )
            ]
        )
        mock_process_library_reports.assert_not_called()

    def test_remove_not_stopped(self, mock_process_library_reports):
        node_identifier = "A"
        resource_ids = ["B"]
        force_flags = []
        self.lib.remote_node.node_remove_remote.side_effect = [
            LibraryError(),
            None,
        ]
        self.report_processor.reports.return_value = [
            UNSTOPPED_RESOURCES_ERROR_REPORT
        ]
        self.lib.remote_node.get_resource_ids.return_value = resource_ids

        self._call_cmd([node_identifier])

        self.report_processor.assert_called_once_with(False)
        self.assert_lib_calls(
            [
                mock.call.remote_node.node_remove_remote(node_identifier, []),
                mock.call.remote_node.get_resource_ids(node_identifier),
                mock.call.resource.stop(resource_ids, force_flags),
                mock.call.cluster.wait_for_pcmk_idle(None),
                mock.call.remote_node.node_remove_remote(
                    node_identifier, force_flags
                ),
            ]
        )
        mock_process_library_reports.assert_not_called()

    def test_remove_more_errors_should_print_reports(
        self, mock_process_library_reports
    ):
        self.lib.remote_node.node_remove_remote.side_effect = [
            LibraryError(),
            None,
        ]
        self.report_processor.reports.return_value = [
            UNSTOPPED_RESOURCES_ERROR_REPORT,
            NODE_NOT_FOUND_ERROR_REPORT,
        ]

        self.assertRaises(LibraryError, lambda: self._call_cmd(["A"]))

        self.report_processor.assert_called_once_with(False)
        self.assert_lib_calls(
            [mock.call.remote_node.node_remove_remote("A", [])]
        )
        mock_process_library_reports.assert_called_once_with(
            [NODE_NOT_FOUND_ERROR_REPORT],
            include_debug=False,
            exit_on_error=False,
        )

    def test_remove_more_errors_reports_already_reported(
        self, mock_process_library_reports
    ):
        self.lib.remote_node.node_remove_remote.side_effect = [
            LibraryError(),
            None,
        ]
        self.report_processor.already_reported_to_console.return_value = True

        self.assertRaises(LibraryError, lambda: self._call_cmd(["A"]))

        self.report_processor.assert_called_once_with(False)
        self.assert_lib_calls(
            [mock.call.remote_node.node_remove_remote("A", [])]
        )
        mock_process_library_reports.assert_not_called()

    def test_remove_more_errors_debug(self, mock_process_library_reports):
        self.lib.remote_node.node_remove_remote.side_effect = [
            LibraryError(),
            None,
        ]
        self.report_processor.reports.return_value = [
            UNSTOPPED_RESOURCES_ERROR_REPORT,
            NODE_NOT_FOUND_ERROR_REPORT,
        ]

        self.assertRaises(
            LibraryError, lambda: self._call_cmd(["A"], {"debug": True})
        )

        self.report_processor.assert_called_once_with(True)
        self.assert_lib_calls(
            [mock.call.remote_node.node_remove_remote("A", [])]
        )
        mock_process_library_reports.assert_called_once_with(
            [NODE_NOT_FOUND_ERROR_REPORT],
            include_debug=True,
            exit_on_error=False,
        )

    def test_mutually_exclusive_options(self, mock_process_library_reports):
        with self.assertRaises(CmdLineInputError) as cm:
            command.node_remove_remote(
                self.lib,
                ["R1"],
                InputModifiers({"-f": "foo", "--no-stop": True}),
            )
        self.assertEqual(
            cm.exception.message, "Only one of '--no-stop', '-f' can be used"
        )
        self.assert_lib_calls([])
        mock_process_library_reports.assert_not_called()


@mock.patch("pcs.cli.cluster.command.process_library_reports")
class NodeRemoveRemote(NodeRemoveRemoteBase, TestCase):
    @mock.patch("pcs.cli.cluster.command.deprecation_warning")
    def test_remove_force(
        self, mock_deprecation_warning, mock_process_library_reports
    ):
        self._call_cmd(["A"], {"force": True})

        mock_deprecation_warning.assert_called_once()
        self.report_processor.assert_not_called()
        self.assert_lib_calls(
            [
                mock.call.remote_node.node_remove_remote(
                    "A", [reports.codes.FORCE]
                )
            ]
        )
        mock_process_library_reports.assert_not_called()


@mock.patch("pcs.cli.cluster.command.process_library_reports")
class NodeRemoveRemoteFuture(NodeRemoveRemoteBase, TestCase):
    def _call_cmd(self, argv, modifiers=None):
        default_modifiers = {"future": True}
        command.node_remove_remote(
            self.lib,
            argv,
            dict_to_modifiers(
                modifiers | default_modifiers
                if modifiers
                else default_modifiers
            ),
        )

    def test_remove_force_forceable_error(self, mock_process_library_reports):
        node_identifier = "A"
        resource_ids = ["B"]
        force_flags = [reports.codes.FORCE]
        self.lib.remote_node.node_remove_remote.side_effect = [
            LibraryError(),
            None,
        ]
        self.lib.remote_node.get_resource_ids.return_value = resource_ids
        self.report_processor.reports.return_value = [FORCEABLE_ERROR_REPORT]

        self._call_cmd([node_identifier], {"force": True})

        self.report_processor.assert_called_once_with(False)
        self.assert_lib_calls(
            [
                mock.call.remote_node.node_remove_remote(node_identifier, []),
                mock.call.remote_node.get_resource_ids(node_identifier),
                mock.call.resource.stop(resource_ids, force_flags),
                mock.call.cluster.wait_for_pcmk_idle(None),
                mock.call.remote_node.node_remove_remote(
                    node_identifier, force_flags
                ),
            ]
        )
        mock_process_library_reports.assert_not_called()

    def test_remove_force_more_errors_not_forceable(
        self, mock_process_library_reports
    ):
        self.lib.remote_node.node_remove_remote.side_effect = [
            LibraryError(),
            None,
        ]
        self.report_processor.reports.return_value = [
            FORCEABLE_ERROR_REPORT,
            NODE_NOT_FOUND_ERROR_REPORT,
        ]

        self.assertRaises(
            LibraryError, lambda: self._call_cmd(["A"], {"force": True})
        )
        self.report_processor.assert_called_once_with(False)
        self.assert_lib_calls(
            [mock.call.remote_node.node_remove_remote("A", [])]
        )
        mock_process_library_reports.assert_called_once_with(
            [
                reports.item.ReportItem.warning(
                    reports.messages.NodeNotFound("A", ["remote"])
                ),
                NODE_NOT_FOUND_ERROR_REPORT,
            ],
            include_debug=False,
            exit_on_error=False,
        )

    def test_remove_with_force_and_skip_offline(
        self, mock_process_library_reports
    ):
        self.report_processor.reports.return_value = [INFO_REPORT]
        self._call_cmd(["A"], {"force": True, "skip-offline": True})

        self.assert_lib_calls(
            [
                mock.call.remote_node.node_remove_remote(
                    "A", [reports.codes.SKIP_OFFLINE_NODES]
                )
            ]
        )
        mock_process_library_reports.assert_called_once_with(
            [INFO_REPORT], include_debug=False
        )


class ClusterRename(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster"])
        self.lib.cluster = mock.Mock(spec_set=["rename"])

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            command.cluster_rename(self.lib, [], dict_to_modifiers({}))
        self.assertIsNone(cm.exception.message)
        self.lib.cluster.rename.assert_not_called()

    def test_too_many_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            command.cluster_rename(self.lib, ["A", "B"], dict_to_modifiers({}))
        self.assertIsNone(cm.exception.message)
        self.lib.cluster.rename.assert_not_called()

    def test_success(self):
        command.cluster_rename(self.lib, ["A"], dict_to_modifiers({}))
        self.lib.cluster.rename.assert_called_once_with("A", [])

    def test_force(self):
        command.cluster_rename(
            self.lib, ["A"], dict_to_modifiers({"force": True})
        )
        self.lib.cluster.rename.assert_called_once_with(
            "A", [report_codes.FORCE]
        )

    def test_skip_offline(self):
        command.cluster_rename(
            self.lib, ["A"], dict_to_modifiers({"skip-offline": True})
        )
        self.lib.cluster.rename.assert_called_once_with(
            "A", [report_codes.SKIP_OFFLINE_NODES]
        )

    def test_all_flags(self):
        command.cluster_rename(
            self.lib,
            ["A"],
            dict_to_modifiers({"force": True, "skip-offline": True}),
        )
        self.lib.cluster.rename.assert_called_once_with(
            "A", [report_codes.FORCE, report_codes.SKIP_OFFLINE_NODES]
        )



from functools import partial
from unittest import TestCase

from pcs.common.host import Destination
from pcs.common.reports import codes as report_codes
from pcs.lib.commands.remote_node import (
    node_remove_guest as node_remove_guest_orig,
)

from pcs_test.tier0.lib.commands.remote_node.fixtures_add import (
    FAIL_HTTP_KWARGS,
    report_manage_services_connection_failed,
)
from pcs_test.tier0.lib.commands.remote_node.fixtures_remove import (
    EnvConfigMixin,
    base_reports_for_host,
    report_authkey_remove_failed,
    report_pcmk_remote_disable_failed,
    report_pcmk_remote_stop_failed,
    report_remove_file_connection_failed,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

REMOTE_HOST = "remote-host"
REMOTE_DEST_LIST = [Destination("remote-addr", 2224)]
NODE_NAME = "node-name"
NODE_DEST_LIST = [Destination("node-addr", 2224)]
VIRTUAL_MACHINE_ID = "virtual_machine_id"


def node_remove_guest(env, node_identifier=REMOTE_HOST, **kwargs):
    node_remove_guest_orig(env, node_identifier, **kwargs)


REPORTS = base_reports_for_host(NODE_NAME)

FIXTURE_RESOURCES = """
    <resources>
        <primitive class="ocf" id="{0}"
            provider="heartbeat" type="VirtualDomain"
        >
            <meta_attributes id="virtual_machine_id-meta_attributes">
                <nvpair id="virtual_machine_id-meta_attributes-remote-addr"
                    name="remote-addr" value="{1}"
                />
                <nvpair id="virtual_machine_id-meta_attributes-remote-node"
                    name="remote-node" value="{2}"
                />
             </meta_attributes>
        </primitive>
    </resources>
""".format(VIRTUAL_MACHINE_ID, REMOTE_HOST, NODE_NAME)

GUEST_NVPAIR_XPATHS = [
    ".//primitive/meta_attributes/nvpair[@name='remote-addr']",
    ".//primitive/meta_attributes/nvpair[@name='remote-node']",
]

get_env_tools = partial(
    get_env_tools, local_extensions={"local": EnvConfigMixin}
)


class RemoveGuest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
            }
        )

    def find_by(self, identifier):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES)
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
        )
        self.config.env.push_cib(remove=GUEST_NVPAIR_XPATHS)
        self.config.runner.pcmk.remove_node(NODE_NAME)

        node_remove_guest(self.env_assist.get_env(), node_identifier=identifier)
        self.env_assist.assert_reports(REPORTS)

    def test_success_base(self):
        self.find_by(REMOTE_HOST)

    def test_can_find_by_node_name(self):
        self.find_by(NODE_NAME)

    def test_can_find_by_resource_id(self):
        self.find_by(VIRTUAL_MACHINE_ID)


class RemoveGuestOthers(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
            }
        )

    def test_success_with_wait(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        wait = 10
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES)
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
        )
        self.config.env.push_cib(remove=GUEST_NVPAIR_XPATHS, wait=wait)
        self.config.runner.pcmk.remove_node(NODE_NAME)

        node_remove_guest(self.env_assist.get_env(), wait=wait)
        self.env_assist.assert_reports(REPORTS)

    def test_can_skip_all_offline(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES)
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST, **FAIL_HTTP_KWARGS
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
            **FAIL_HTTP_KWARGS,
        )
        self.config.env.push_cib(remove=GUEST_NVPAIR_XPATHS)
        self.config.runner.pcmk.remove_node(NODE_NAME)

        node_remove_guest(self.env_assist.get_env(), skip_offline_nodes=True)
        my_reports = REPORTS.copy()
        my_reports.replace(
            "pcmk_remote_disable_success",
            report_manage_services_connection_failed(NODE_NAME).to_warn(),
        )
        my_reports.remove("pcmk_remote_stop_success")
        my_reports.replace(
            "authkey_remove_success",
            report_remove_file_connection_failed(NODE_NAME).to_warn(),
        )
        self.env_assist.assert_reports(my_reports)

    def test_fail_when_identifier_not_found(self):
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES)
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_guest(
                self.env_assist.get_env(), node_identifier="NONEXISTENT"
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    node="NONEXISTENT",
                    searched_types=["guest"],
                )
            ]
        )


class MultipleResults(TestCase):
    fixture_multi_resources = """
        <resources>
            <primitive class="ocf" id="{0}"
                provider="heartbeat" type="VirtualDomain"
            >
                <meta_attributes id="A-M">
                    <nvpair id="A-M-RAddr" name="remote-addr" value="{1}"/>
                    <nvpair id="A-M-RNode" name="remote-node" value="{2}"/>
                 </meta_attributes>
            </primitive>

            <primitive class="ocf" id="{1}"
                provider="heartbeat" type="VirtualDomain"
            >
                <meta_attributes id="B-M">
                    <nvpair id="B-M-RAddr" name="remote-addr" value="{3}"/>
                    <nvpair id="B-M-RNode" name="remote-node" value="{4}"/>
                 </meta_attributes>
            </primitive>

            <primitive class="ocf" id="C"
                provider="heartbeat" type="VirtualDomain"
            >
                <meta_attributes id="C-M">
                    <nvpair id="C-M-RAddr" name="remote-addr" value="{2}"/>
                    <nvpair id="C-M-RNode" name="remote-node" value="{1}"/>
                 </meta_attributes>
            </primitive>
        </resources>
    """.format(VIRTUAL_MACHINE_ID, REMOTE_HOST, NODE_NAME, "B-HOST", "B-NAME")

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(resources=self.fixture_multi_resources)
        self.report_multiple_results = fixture.error(
            report_codes.MULTIPLE_RESULTS_FOUND,
            force_code=report_codes.FORCE,
            result_identifier_list=[
                VIRTUAL_MACHINE_ID,
                REMOTE_HOST,
                "C",
            ],
            result_type="resource",
            search_description=REMOTE_HOST,
        )
        self.dest_list_b = [Destination("B-ADDR", 2224)]
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
                REMOTE_HOST: REMOTE_DEST_LIST,
                "B-NAME": self.dest_list_b,
            }
        )

    def test_fail(self):
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_guest(
                self.env_assist.get_env(), node_identifier=REMOTE_HOST
            )
        )
        self.env_assist.assert_reports([self.report_multiple_results])

    def test_force(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.config.local.destroy_pacemaker_remote(
            communication_list=[
                dict(label="B-NAME", dest_list=self.dest_list_b),
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST),
                dict(label=REMOTE_HOST, dest_list=REMOTE_DEST_LIST),
            ],
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label="B-NAME", dest_list=self.dest_list_b),
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST),
                dict(label=REMOTE_HOST, dest_list=REMOTE_DEST_LIST),
            ],
        )
        self.config.env.push_cib(
            remove=[
                ".//meta_attributes[@id='A-M']/nvpair[@name='remote-addr']",
                ".//meta_attributes[@id='A-M']/nvpair[@name='remote-node']",
                ".//meta_attributes[@id='B-M']/nvpair[@name='remote-addr']",
                ".//meta_attributes[@id='B-M']/nvpair[@name='remote-node']",
                ".//meta_attributes[@id='C-M']/nvpair[@name='remote-addr']",
                ".//meta_attributes[@id='C-M']/nvpair[@name='remote-node']",
            ]
        )
        self.config.runner.pcmk.remove_node(
            "B-NAME", name="runner.pcmk.remove_node3"
        )
        self.config.runner.pcmk.remove_node(
            NODE_NAME, name="runner.pcmk.remove_node2"
        )
        self.config.runner.pcmk.remove_node(REMOTE_HOST)

        node_remove_guest(
            self.env_assist.get_env(),
            node_identifier=REMOTE_HOST,
            allow_remove_multiple_nodes=True,
        )
        my_reports = REPORTS.copy()
        my_reports.replace(
            "pcmk_remote_disable_stop_started",
            REPORTS["pcmk_remote_disable_stop_started"].adapt(
                node_list=["B-NAME", NODE_NAME, REMOTE_HOST]
            ),
        )
        my_reports.append(
            REPORTS["pcmk_remote_disable_success"].adapt(node="B-NAME")
        )
        my_reports.append(
            REPORTS["pcmk_remote_stop_success"].adapt(node="B-NAME")
        )
        my_reports.append(
            REPORTS["pcmk_remote_disable_success"].adapt(node=REMOTE_HOST)
        )
        my_reports.append(
            REPORTS["pcmk_remote_stop_success"].adapt(node=REMOTE_HOST)
        )
        my_reports.replace(
            "authkey_remove_started",
            REPORTS["authkey_remove_started"].adapt(
                node_list=["B-NAME", NODE_NAME, REMOTE_HOST]
            ),
        )
        my_reports.append(
            REPORTS["authkey_remove_success"].adapt(node="B-NAME")
        )
        my_reports.append(
            REPORTS["authkey_remove_success"].adapt(node=REMOTE_HOST)
        )
        my_reports.append(self.report_multiple_results.to_warn())
        self.env_assist.assert_reports(my_reports)


class AuthkeyRemove(TestCase):
    def setUp(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES)
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST
        )
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
            }
        )

    def test_fails_when_offline(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
            **FAIL_HTTP_KWARGS,
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_guest(self.env_assist.get_env())
        )
        my_reports = REPORTS.copy()
        my_reports.replace(
            "authkey_remove_success",
            report_remove_file_connection_failed(NODE_NAME),
        )
        self.env_assist.assert_reports(my_reports)

    def test_fails_when_remotely_fails(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
            result={
                "code": "unexpected",
                "message": "Access denied",
            },
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_guest(self.env_assist.get_env())
        )
        my_reports = REPORTS.copy()
        my_reports.replace(
            "authkey_remove_success",
            report_authkey_remove_failed(NODE_NAME),
        )
        self.env_assist.assert_reports(my_reports)

    def test_forceable_when_remotely_fail(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
            result={
                "code": "unexpected",
                "message": "Access denied",
            },
        )
        self.config.env.push_cib(remove=GUEST_NVPAIR_XPATHS)
        self.config.runner.pcmk.remove_node(NODE_NAME)

        node_remove_guest(
            self.env_assist.get_env(), allow_pacemaker_remote_service_fail=True
        )
        my_reports = REPORTS.copy()
        my_reports.replace(
            "authkey_remove_success",
            report_authkey_remove_failed(NODE_NAME).to_warn(),
        )
        self.env_assist.assert_reports(my_reports)


class PcmkRemoteServiceDestroy(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
            }
        )
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES)

    def test_fails_when_offline(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST, **FAIL_HTTP_KWARGS
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_guest(self.env_assist.get_env())
        )
        my_reports = REPORTS[:"pcmk_remote_disable_success"]
        my_reports.append(report_manage_services_connection_failed(NODE_NAME))
        self.env_assist.assert_reports(my_reports)

    def test_fails_when_remotely_fails(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME,
            dest_list=NODE_DEST_LIST,
            result={
                "code": "fail",
                "message": "Action failed",
            },
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_guest(self.env_assist.get_env())
        )
        my_reports = REPORTS[:"pcmk_remote_disable_success"]
        my_reports.append(report_pcmk_remote_disable_failed(NODE_NAME))
        my_reports.append(report_pcmk_remote_stop_failed(NODE_NAME))
        self.env_assist.assert_reports(my_reports)

    def test_forceable_when_remotely_fail(self):
        # Instance of 'Config' has no 'local' member
        # pylint: disable=no-member
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME,
            dest_list=NODE_DEST_LIST,
            result={
                "code": "fail",
                "message": "Action failed",
            },
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
        )
        self.config.env.push_cib(remove=GUEST_NVPAIR_XPATHS)
        self.config.runner.pcmk.remove_node(NODE_NAME)

        node_remove_guest(
            self.env_assist.get_env(), allow_pacemaker_remote_service_fail=True
        )
        my_reports = REPORTS.copy()
        my_reports.replace(
            "pcmk_remote_disable_success",
            report_pcmk_remote_disable_failed(NODE_NAME).to_warn(),
        )
        my_reports.replace(
            "pcmk_remote_stop_success",
            report_pcmk_remote_stop_failed(NODE_NAME).to_warn(),
        )
        self.env_assist.assert_reports(my_reports)



import os.path
import signal
import subprocess
from typing import (
    Mapping,
    Optional,
)

from pcs.common.types import StringSequence

from pcs_test import TEST_ROOT

__pcs_location = os.path.join(TEST_ROOT, "pcs_for_tests")
# this can be changed from suite.py
test_installed = False


class PcsRunner:
    def __init__(
        self,
        cib_file: str,
        corosync_conf_opt: Optional[str] = None,
        mock_settings: Optional[Mapping[str, str]] = None,
    ):
        self.cib_file = cib_file
        self.corosync_conf_opt = corosync_conf_opt
        self.mock_settings = mock_settings

    def run(self, args: StringSequence) -> tuple[str, str, int]:
        return pcs(
            self.cib_file,
            args,
            corosync_conf_opt=self.corosync_conf_opt,
            mock_settings=self.mock_settings,
        )


def pcs(
    cib_file: Optional[str],
    args: StringSequence,
    corosync_conf_opt: Optional[str] = None,
    mock_settings: Optional[Mapping[str, str]] = None,
) -> tuple[str, str, int]:
    if mock_settings is None:
        mock_settings = {}

    env_mock_settings_prefix = "PCS.SETTINGS."
    env = {
        "{}{}".format(env_mock_settings_prefix, option): value
        for option, value in mock_settings.items()
    }
    if test_installed:
        env["PCS_TEST.TEST_INSTALLED"] = "1"

    cmd = [__pcs_location]
    if cib_file:
        cmd.extend(["-f", cib_file])
    if corosync_conf_opt:
        cmd.extend(["--corosync_conf", corosync_conf_opt])
    cmd += args

    return _run(cmd, env_extend=env)


def _run(
    args: StringSequence, env_extend: Optional[Mapping[str, str]] = None
) -> tuple[str, str, int]:
    env_vars = {"LC_ALL": "C"}
    env_vars.update(dict(env_extend) if env_extend else {})

    # pylint: disable=subprocess-popen-preexec-fn
    with subprocess.Popen(
        args,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        preexec_fn=(lambda: signal.signal(signal.SIGPIPE, signal.SIG_DFL)),  # noqa: PLW1509
        close_fds=True,
        shell=False,
        env=env_vars,
        # decodes newlines and in python3 also converts bytes to str
        universal_newlines=True,
    ) as process:
        stdout, stderr = process.communicate()
        retval = process.returncode

    return stdout, stderr, retval
